from glycresoft.task import TaskBase

from .chromatogram import (get_chromatogram, mask_subsequence)
from .index import ChromatogramFilter


class MassShiftTreePruner(TaskBase):
    def __init__(self, solutions, score_margin=2.5, ratio_threshold=1.5, trivial_abundance_delta_ratio=0.01):
        self.solutions = solutions
        self.score_margin = score_margin
        self.ratio_threshold = ratio_threshold
        self.trivial_abundance_delta_ratio = trivial_abundance_delta_ratio

        self.key_map = self.solutions._build_key_map()
        self.updated = set()

    def check_close_score_abundance_ratio(self, case, owner_item):
        component_signal = case.total_signal
        complement_signal = owner_item.total_signal - component_signal
        signal_ratio = complement_signal / component_signal
        # The owner is more abundant than used-as-mass_shift-case
        return (signal_ratio < self.ratio_threshold)

    def create_masked_chromatogram(self, owner_item, case):
        new_masked = mask_subsequence(get_chromatogram(owner_item), get_chromatogram(case))
        new_masked.created_at = "prune_bad_mass_shift_branches"
        new_masked.score = owner_item.score
        return new_masked

    def handle_chromatogram(self, case):
        if case.used_as_mass_shift:
            keepers = []
            for owning_key, mass_shift in case.used_as_mass_shift:
                owner = self.key_map.get(owning_key)
                if owner is None:
                    continue
                owner_item = owner.find_overlap(case)
                if owner_item is None:
                    continue
                scores_close = abs(case.score - owner_item.score) < self.score_margin
                mass_shifted_case_stronger = case.score > owner_item.score
                # If the owning group is lower scoring, but the scores are close
                if mass_shifted_case_stronger and scores_close:
                    mass_shifted_case_stronger = self.check_close_score_abundance_ratio(case, owner_item)
                # If the scores are close, but the owning group is less abundant,
                # e.g. more mass shift groups or mass accuracy prevents propagation
                # of mass shifts
                elif scores_close and (owner_item.total_signal / case.total_signal) < 1:
                    mass_shifted_case_stronger = True
                independent = owner_item.total_signal - case.total_signal
                if (independent / case.total_signal) < self.trivial_abundance_delta_ratio:
                    mass_shifted_case_stronger = True
                if mass_shifted_case_stronger:
                    new_masked = self.create_masked_chromatogram(owner_item, case)
                    if len(new_masked) != 0:
                        owner.replace(owner_item, new_masked)
                    self.updated.add(owning_key)
                else:
                    keepers.append((owning_key, mass_shift))
            case.chromatogram.used_as_mass_shift = keepers

    def prune_branches(self):
        n = len(self.solutions)
        for i, case in enumerate(self.solutions):
            if len(case.used_as_mass_shift) > 2 ** 6:
                self.log("Pruning %r with %d mass shift branches" % (case, len(case.used_as_mass_shift)))
            if i % 100 == 0 and i > 0:
                self.log("... %d chromatograms reduced (%0.2f%%)" % (i, i / float(n) * 100.0))
            self.handle_chromatogram(case)
        out = [s.chromatogram for k in (set(self.key_map) - self.updated) for s in self.key_map[k]]
        out.extend(s for k in self.updated for s in self.key_map[k])
        out = ChromatogramFilter(out)
        return out

    @classmethod
    def prune_bad_mass_shift_branches(cls, solutions, score_margin=2.5, ratio_threshold=1.5,
                                      trivial_abundance_delta_ratio=0.01):
        inst = cls(solutions, score_margin, ratio_threshold, trivial_abundance_delta_ratio)
        return inst.prune_branches()


prune_bad_mass_shift_branches = MassShiftTreePruner.prune_bad_mass_shift_branches



from .scoring import (
    BinomialSpectrumMatcher, TargetDecoyAnalyzer,
    CoverageWeightedBinomialScorer, LogIntensityScorer)

from .matcher import GlycopeptideMatcher

from .glycopeptide_matcher import (
    GlycopeptideDatabaseSearchIdentifier,
    ExclusiveGlycopeptideDatabaseSearchComparer,
    GlycopeptideDatabaseSearchComparer)


from .dynamic_generation import (
    MultipartGlycopeptideIdentifier, PeptideDatabaseProxyLoader,
    make_memory_database_proxy_resolver)


__all__ = [
    "BinomialSpectrumMatcher", "TargetDecoyAnalyzer",
    "CoverageWeightedBinomialScorer", "LogIntensityScorer",

    "GlycopeptideDatabaseSearchIdentifier",
    "ExclusiveGlycopeptideDatabaseSearchComparer",
    "GlycopeptideDatabaseSearchComparer",
    "GlycopeptideMatcher",

    "MultipartGlycopeptideIdentifier", "PeptideDatabaseProxyLoader",
    "make_memory_database_proxy_resolver",
]



from collections import defaultdict

from glypy.io import iupac, glycoct
from glypy.structure.glycan_composition import HashableGlycanComposition, FrozenGlycanComposition
from glypy.enzyme import (
    make_n_glycan_pathway, make_mucin_type_o_glycan_pathway,
    MultiprocessingGlycome, Glycosylase, Glycosyltransferase,
    EnzymeGraph, GlycanCompositionEnzymeGraph, _enzyme_graph_inner)

from glycresoft.task import TaskBase, log_handle

from glycresoft.database.builder.glycan.glycan_source import (
    GlycanHypothesisSerializerBase, GlycanTransformer,
    DBGlycanComposition, formula, GlycanCompositionToClass,
    GlycanTypes)

from glycresoft.structure import KeyTransformingDecoratorDict


def key_transform(name):
    return str(name).lower().replace(" ", '-')


synthesis_register = KeyTransformingDecoratorDict(key_transform)


class MultiprocessingGlycomeTask(MultiprocessingGlycome, TaskBase):
    def _log(self, message):
        log_handle.log(message)

    def log_generation_chunk(self, i, chunks, current_generation):
        self._log(".... Task %d/%d finished (%d items generated)" % (
            i, len(chunks), len(current_generation)))


class GlycanSynthesis(TaskBase):
    glycan_classification = None

    def __init__(self, glycosylases=None, glycosyltransferases=None, seeds=None, limits=None,
                 convert_to_composition=True, n_processes=5):
        self.glycosylases = glycosylases or {}
        self.glycosyltransferases = glycosyltransferases or {}
        self.seeds = seeds or []
        self.limits = limits or []
        self.convert_to_composition = convert_to_composition
        self.n_processes = n_processes

    def remove_enzyme(self, enzyme_name):
        if enzyme_name in self.glycosylases:
            return self.glycosylases.pop(enzyme_name)
        elif enzyme_name in self.glycosyltransferases:
            return self.glycosyltransferases.pop(enzyme_name)
        else:
            raise KeyError(enzyme_name)

    def add_enzyme(self, enzyme_name, enzyme):
        if isinstance(enzyme, Glycosylase):
            self.glycosylases[enzyme_name] = enzyme
        elif isinstance(enzyme, Glycosyltransferase):
            self.glycosyltransferases[enzyme_name] = enzyme
        else:
            raise TypeError("Don't know where to put object of type %r" % type(enzyme))

    def add_limit(self, limit):
        self.limits.append(limit)

    def add_seed(self, structure):
        if structure in self.seeds:
            return
        self.seeds.append(structure)

    def build_glycome(self):
        glycome = MultiprocessingGlycomeTask(
            self.glycosylases, self.glycosyltransferases,
            self.seeds, track_generations=False,
            limits=self.limits, processes=self.n_processes)
        return glycome

    def convert_enzyme_graph_composition(self, glycome):
        self.log("Converting Enzyme Graph into Glycan Set")
        glycans = set()
        glycans.update(glycome.enzyme_graph)
        for i, v in enumerate(glycome.enzyme_graph.values()):
            if i and i % 100000 == 0:
                self.log(".... %d Glycans In Set" % (len(glycans)))
            glycans.update(v)
        self.log(".... %d Glycans In Set" % (len(glycans)))

        composition_graph = defaultdict(_enzyme_graph_inner)
        compositions = set()

        cache = StructureConverter()

        i = 0
        for s in glycans:
            i += 1
            gc = cache[s]
            for child, enz in glycome.enzyme_graph[s].items():
                composition_graph[gc][cache[child]].update(enz)
            if i % 1000 == 0:
                self.log(".... Converted %d Compositions (%d/%d Structures, %0.2f%%)" % (
                    len(compositions), i, len(glycans), float(i) / len(glycans) * 100.0))
            compositions.add(gc)
        return compositions, composition_graph

    def extract_structures(self, glycome):
        self.log("Converting Enzyme Graph into Glycan Set")
        solutions = list()
        for i, structure in enumerate(glycome.seen):
            if i and i % 10000 == 0:
                self.log(".... %d Glycans Extracted" % (i,))
            solutions.append(structure)
        return solutions, glycome.enzyme_graph

    def run(self):
        logger = self.ipc_logger()
        glycome = self.build_glycome()
        old_logger = glycome._log
        glycome._log = logger.handler
        for i, gen in enumerate(glycome.run()):
            self.log(".... Generation %d: %d Structures" % (i, len(gen)))
        self.glycome = glycome
        logger.stop()
        glycome._log = old_logger
        if self.convert_to_composition:
            compositions, composition_enzyme_graph = self.convert_enzyme_graph_composition(glycome)
            return compositions, composition_enzyme_graph
        else:
            structures, enzyme_graph = self.extract_structures(glycome)
            return structures, enzyme_graph


class Limiter(object):
    def __init__(self, max_nodes=26, max_mass=5500.0):
        self.max_mass = max_mass
        self.max_nodes = max_nodes

    def __call__(self, x):
        return len(x) < self.max_nodes and x.mass() < self.max_mass


GlycanSynthesis.size_limiter_type = Limiter


@synthesis_register("n-glycan")
@synthesis_register("mammalian-n-glycan")
class NGlycanSynthesis(GlycanSynthesis):

    glycan_classification = GlycanTypes.n_glycan

    def _get_initial_components(self):
        glycosidases, glycosyltransferases, seeds = make_n_glycan_pathway()
        glycosyltransferases.pop('siat2_3')

        child = iupac.loads("a-D-Neup5Gc")
        parent = iupac.loads("b-D-Galp-(1-4)-b-D-Glcp2NAc")
        siagct2_6 = Glycosyltransferase(6, 2, parent, child, parent_node_id=3)

        glycosyltransferases['siagct2_6'] = siagct2_6
        return glycosidases, glycosyltransferases, seeds

    def __init__(self, glycosylases=None, glycosyltransferases=None, seeds=None, limits=None,
                 convert_to_composition=True, n_processes=5):
        sylases, transferases, more_seeds = self._get_initial_components()
        sylases.update(glycosylases or {})
        transferases.update(glycosyltransferases or {})
        more_seeds.extend(seeds or [])
        super(NGlycanSynthesis, self).__init__(sylases, transferases, more_seeds,
                                               limits, convert_to_composition, n_processes)


@synthesis_register("human-n-glycan")
class HumanNGlycanSynthesis(NGlycanSynthesis):
    def _get_initial_components(self):
        glycosidases, glycosyltransferases, seeds = super(
            HumanNGlycanSynthesis, self)._get_initial_components()
        glycosyltransferases.pop('siagct2_6')
        glycosyltransferases.pop('agal13galt')
        glycosyltransferases.pop('gntE')
        return glycosidases, glycosyltransferases, seeds


@synthesis_register("mucin-o-glycan")
@synthesis_register("mammalian-mucin-o-glycan")
class MucinOGlycanSynthesis(GlycanSynthesis):

    glycan_classification = GlycanTypes.o_glycan

    def _get_initial_components(self):
        glycosidases, glycosyltransferases, seeds = make_mucin_type_o_glycan_pathway()
        parent = iupac.loads("a-D-Galp2NAc")
        child = iupac.loads("a-D-Neup5Gc")
        sgt6gal1 = Glycosyltransferase(6, 2, parent, child, terminal=False)
        glycosyltransferases['sgt6gal1'] = sgt6gal1

        parent = iupac.loads("b-D-Galp-(1-3)-a-D-Galp2NAc")
        child = iupac.loads("a-D-Neup5Gc")
        sgt3gal2 = Glycosyltransferase(3, 2, parent, child, parent_node_id=3)
        glycosyltransferases['sgt3gal2'] = sgt3gal2

        parent = iupac.loads("b-D-Galp-(1-3)-a-D-Galp2NAc")
        child = iupac.loads("a-D-Neup5Gc")
        sgt6gal2 = Glycosyltransferase(6, 2, parent, child, parent_node_id=3)
        glycosyltransferases['sgt6gal2'] = sgt6gal2
        return glycosidases, glycosyltransferases, seeds

    def __init__(self, glycosylases=None, glycosyltransferases=None, seeds=None, limits=None,
                 convert_to_composition=True, n_processes=5):
        sylases, transferases, more_seeds = self._get_initial_components()
        sylases.update(glycosylases or {})
        transferases.update(glycosyltransferases or {})
        more_seeds.extend(seeds or [])
        super(MucinOGlycanSynthesis, self).__init__(sylases, transferases, more_seeds,
                                                    limits, convert_to_composition, n_processes)


@synthesis_register("human-mucin-o-glycan")
class HumanMucinOGlycanSynthesis(MucinOGlycanSynthesis):
    def _get_initial_components(self):
        glycosidases, glycosyltransferases, seeds = super(HumanMucinOGlycanSynthesis)._get_initial_components()
        glycosyltransferases.pop("sgt6gal1")
        glycosyltransferases.pop("sgt3gal2")
        glycosyltransferases.pop("sgt6gal2")
        return glycosidases, glycosyltransferases, seeds


class StructureConverter(object):
    def __init__(self):
        self.cache = dict()

    def convert(self, structure_text):
        if structure_text in self.cache:
            return self.cache[structure_text]
        structure = glycoct.loads(structure_text)
        gc = HashableGlycanComposition.from_glycan(structure).thaw()
        gc.drop_stems()
        gc.drop_configurations()
        gc.drop_positions()
        gc = HashableGlycanComposition(gc)
        self.cache[structure_text] = gc
        return gc

    def __getitem__(self, structure_text):
        return self.convert(structure_text)

    def __repr__(self):
        return "%s(%d)" % (self.__class__.__name__, len(self.cache))


class AdaptExistingGlycanGraph(TaskBase):
    def __init__(self, graph, enzymes_to_remove):
        self.graph = graph
        self.enzymes_to_remove = set(enzymes_to_remove)
        self.enzymes_available = set(self.graph.enzymes())
        if (self.enzymes_to_remove - self.enzymes_available):
            raise ValueError("Required enzymes %r not found" % (
                self.enzymes_to_remove - self.enzymes_available,))

    def remove_enzymes(self):
        enz_to_remove = self.enzymes_to_remove
        for enz in enz_to_remove:
            self.log(".... Removing Enzyme %s" % (enz,))
            self.graph.remove_enzyme(enz)
            for entity in (self.graph.parentless() - self.graph.seeds):
                self.graph.remove(entity)

    def run(self):
        self.log("Adapting Enzyme Graph with %d nodes and %d edges" % (
            self.graph.node_count(), self.graph.edge_count()))
        self.remove_enzymes()
        self.log("After Adaption, Graph has %d nodes and %d edges" % (
            self.graph.node_count(), self.graph.edge_count()))


class ExistingGraphGlycanHypothesisSerializer(GlycanHypothesisSerializerBase):
    def __init__(self, enzyme_graph, database_connection, enzymes_to_remove=None, reduction=None,
                 derivatization=None, hypothesis_name=None, glycan_classification=None):
        if enzymes_to_remove is None:
            enzymes_to_remove = set()

        GlycanHypothesisSerializerBase.__init__(self, database_connection, hypothesis_name)

        self.enzyme_graph = enzyme_graph
        self.enzymes_to_remove = set(enzymes_to_remove)
        self.glycan_classification = glycan_classification

        self.reduction = reduction
        self.derivatization = derivatization

        self.loader = None
        self.transformer = None

    def build_glycan_compositions(self):
        adapter = AdaptExistingGlycanGraph(self.enzyme_graph, self.enzymes_to_remove)
        adapter.start()
        components = adapter.graph.nodes()
        for component in components:
            if isinstance(component, FrozenGlycanComposition):
                component = component.thaw()
            yield component, [self.glycan_classification]

    def make_pipeline(self):
        self.loader = self.build_glycan_compositions()
        self.transformer = GlycanTransformer(self.loader, self.reduction, self.derivatization)

    def run(self):
        self.make_pipeline()
        structure_class_lookup = self.structure_class_loader

        acc = []
        counter = 0
        for composition, structure_classes in self.transformer:
            mass = composition.mass()
            composition_string = composition.serialize()
            formula_string = formula(composition.total_composition())
            inst = DBGlycanComposition(
                calculated_mass=mass, formula=formula_string,
                composition=composition_string,
                hypothesis_id=self.hypothesis_id)
            self.session.add(inst)
            self.session.flush()
            counter += 1
            for structure_class in structure_classes:
                structure_class = structure_class_lookup[structure_class]
                acc.append(dict(glycan_id=inst.id, class_id=structure_class.id))
                if len(acc) % 100 == 0:
                    self.session.execute(GlycanCompositionToClass.insert(), acc)
                    acc = []
        if acc:
            self.session.execute(GlycanCompositionToClass.insert(), acc)
            acc = []
        self.session.commit()
        self.log("Generated %d glycan compositions" % counter)


class SynthesisGlycanHypothesisSerializer(GlycanHypothesisSerializerBase):
    def __init__(self, glycome, database_connection, reduction=None,
                 derivatization=None, hypothesis_name=None, glycan_classification=None):
        if glycan_classification is None:
            glycan_classification = glycome.glycan_classification

        GlycanHypothesisSerializerBase.__init__(self, database_connection, hypothesis_name)

        self.glycome = glycome
        self.glycan_classification = glycan_classification

        self.reduction = reduction
        self.derivatization = derivatization

        self.loader = None
        self.transformer = None

    def build_glycan_compositions(self):
        components, enzyme_graph = self.glycome.start()
        for component in components:
            yield component, [self.glycan_classification]

    def make_pipeline(self):
        self.loader = self.build_glycan_compositions()
        self.transformer = GlycanTransformer(self.loader, self.reduction, self.derivatization)

    def run(self):
        self.make_pipeline()
        structure_class_lookup = self.structure_class_loader

        acc = []
        counter = 0
        for composition, structure_classes in self.transformer:
            mass = composition.mass()
            composition_string = composition.serialize()
            formula_string = formula(composition.total_composition())
            inst = DBGlycanComposition(
                calculated_mass=mass, formula=formula_string,
                composition=composition_string,
                hypothesis_id=self.hypothesis_id)
            if (counter + 1) % 100 == 0:
                self.log("Stored %d glycan compositions" % counter)
            self.session.add(inst)
            self.session.flush()
            counter += 1
            for structure_class in structure_classes:
                structure_class = structure_class_lookup[structure_class]
                acc.append(dict(glycan_id=inst.id, class_id=structure_class.id))
                if len(acc) % 100 == 0:
                    self.session.execute(GlycanCompositionToClass.insert(), acc)
                    acc = []
        if acc:
            self.session.execute(GlycanCompositionToClass.insert(), acc)
            acc = []
        self.session.commit()
        self.log("Stored %d glycan compositions" % counter)



from collections import OrderedDict
from matplotlib import pyplot as plt

from .chromatogram_artist import (
    SmoothingChromatogramArtist, AbundantLabeler,
    NGlycanLabelProducer, n_glycan_colorizer)

from .entity_bar_chart import AggregatedAbundanceArtist, BundledGlycanComposition
from .utils import figax


class GlycanChromatographySummaryGraphBuilder(object):
    def __init__(self, solutions):
        self.solutions = solutions

    def chromatograms(self, min_score=0.4, min_signal=0.2, colorizer=None, total_ion_chromatogram=None,
                      base_peak_chromatogram=None, ax=None):
        if ax is None:
            ax = figax()
        monosaccharides = set()

        for sol in self.solutions:
            if sol.glycan_composition:
                monosaccharides.update(map(str, sol.glycan_composition))

        label_abundant = AbundantLabeler(
            NGlycanLabelProducer(monosaccharides),
            max(sol.total_signal for sol in self.solutions if sol.score > min_score) * min_signal)

        if colorizer is None:
            colorizer = n_glycan_colorizer

        results = [sol for sol in self.solutions if sol.score > min_score and not sol.used_as_mass_shift]
        chrom = SmoothingChromatogramArtist(
            results, ax=ax, colorizer=colorizer).draw(label_function=label_abundant)

        if total_ion_chromatogram is not None:
            rt, intens = total_ion_chromatogram.as_arrays()
            chrom.draw_generic_chromatogram(
                "TIC", rt, intens, 'blue')
            chrom.ax.set_ylim(0, max(intens) * 1.1)

        if base_peak_chromatogram is not None:
            rt, intens = base_peak_chromatogram.as_arrays()
            chrom.draw_generic_chromatogram(
                "BPC", rt, intens, 'green')
        return chrom

    def aggregated_abundance(self, min_score=0.4, ax=None):
        if ax is None:
            ax = figax()
        agg = AggregatedAbundanceArtist(
            BundledGlycanComposition.aggregate([
                sol for sol in self.solutions if (sol.score > min_score and
                                                  sol.glycan_composition is not None and
                                                  not sol.used_as_mass_shift)]),
            ax=ax)
        if len(agg) == 0:
            ax = agg.ax
            ax.text(0.5, 0.5, "No Entities Matched", ha='center')
            ax.set_axis_off()
        else:
            agg.draw()
        return agg

    def draw(self, min_score=0.4, min_signal=0.2, colorizer=None, total_ion_chromatogram=None,
             base_peak_chromatogram=None):
        chrom = self.chromatograms(min_score, min_signal, colorizer,
                                   total_ion_chromatogram, base_peak_chromatogram)
        agg = self.aggregated_abundance(min_score)
        return chrom, agg


def breaklines(cases):
    counts = OrderedDict()
    counter = 0
    ms2_score = 0
    cases = sorted(cases, key=lambda x: x.ms2_score, reverse=True)
    i = 0
    for case in cases:
        i += 1
        if case.ms2_score == ms2_score:
            counter += 1
        else:
            counts[ms2_score] = counter
            ms2_score = case.ms2_score
            counter += 1
    if counts[0] == 0:
        counts.pop(0)
    return counts


def plot_tapering(cases, threshold=0.05, ax=None, **kwargs):
    plot_kwargs = {
        "alpha": 0.5,
        "lw": 2
    }
    plot_kwargs.update(kwargs)
    counts = breaklines(cases)
    if ax is None:
        fig, ax = plt.subplots(1)

    score_at_threshold = float('inf')
    for t in cases:
        if t.q_value < 0.05:
            if t.score < score_at_threshold:
                score_at_threshold = t.score

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.xaxis.tick_bottom()
    ax.yaxis.tick_left()
    ax.plot(*zip(*counts.items()), **plot_kwargs)
    ax.set_xlim(*(ax.get_xlim()[::-1]))

    xlim = ax.get_xlim()
    ax.hlines(counts[score_at_threshold], max(xlim), 0, linestyles='--')

    ax.set_xlabel("PSM Score Threshold", fontsize=18)
    ax.set_ylabel("# of PSMS < Threshold", fontsize=18)
    return ax



cimport cython
from cpython.object cimport PyObject
from cpython.ref cimport Py_INCREF
from cpython.dict cimport PyDict_GetItem, PyDict_SetItem
from cpython.tuple cimport PyTuple_GET_SIZE, PyTuple_GET_ITEM, PyTuple_SET_ITEM, PyTuple_New
from cpython.list cimport PyList_GET_SIZE, PyList_GET_ITEM, PyList_SET_ITEM, PyList_New

from numpy cimport npy_uint32 as uint32_t, npy_uint64 as uint64_t, npy_float64 as float64_t
cimport numpy as np

np.import_array()

from glypy.utils.cenum cimport EnumMeta, EnumValue
from glycopeptidepy._c.structure.fragment cimport StubFragment


@cython.freelist(1000000)
cdef class GlycopeptideDatabaseRecord(object):
    def __init__(self, id, calculated_mass, glycopeptide_sequence, protein_id,
                 start_position, end_position, peptide_mass, hypothesis_id):
        self.id = id
        self.calculated_mass = calculated_mass
        self.glycopeptide_sequence = glycopeptide_sequence
        self.protein_id = protein_id
        self.start_position = start_position
        self.end_position = end_position
        self.peptide_mass = peptide_mass
        self.hypothesis_id = hypothesis_id

    def __repr__(self):
        template = (
            "{self.__class__.__name__}(id={self.id}, calculated_mass={self.calculated_mass}, "
            "glycopeptide_sequence={self.glycopeptide_sequence}, protein_id={self.protein_id}, "
            "start_position={self.start_position}, end_position={self.end_position}, "
            "peptide_mass={self.peptide_mass}, hypothesis_id={self.hypothesis_id}, ")
        return template.format(self=self)

    def __reduce__(self):
        return self.__class__, (self.id, self.calculated_mass, self.glycopeptide_sequence, self.protein_id,
                                self.start_position, self.end_position, self.peptide_mass, self.hypothesis_id)


@cython.freelist(10000000)
cdef class glycopeptide_key(object):
    def __init__(self, start_position, end_position, peptide_id, protein_id, hypothesis_id,
                 glycan_combination_id, structure_type, site_combination_index):
        self.start_position = start_position
        self.end_position = end_position
        self.peptide_id = peptide_id
        self.protein_id = protein_id
        self.hypothesis_id = hypothesis_id
        self.glycan_combination_id = glycan_combination_id
        self.structure_type = structure_type
        self.site_combination_index = site_combination_index
        self._hash = hash((
            start_position, end_position, peptide_id, protein_id, hypothesis_id,
            glycan_combination_id, structure_type, site_combination_index))

    @staticmethod
    cdef glycopeptide_key _create(uint32_t start_position, uint32_t end_position, uint64_t peptide_id,
                                  uint32_t protein_id, uint32_t hypothesis_id, uint64_t glycan_combination_id,
                                  object structure_type, uint64_t site_combination_index, Py_hash_t _hash=0):
        cdef glycopeptide_key self = glycopeptide_key.__new__(glycopeptide_key)
        self.start_position = start_position
        self.end_position = end_position
        self.peptide_id = peptide_id
        self.protein_id = protein_id
        self.hypothesis_id = hypothesis_id
        self.glycan_combination_id = glycan_combination_id
        self.structure_type = structure_type
        self.site_combination_index = site_combination_index
        if _hash == 0:
            self._hash = hash((
                start_position, end_position, peptide_id, protein_id, hypothesis_id,
                glycan_combination_id, structure_type, site_combination_index))
        else:
            self._hash = _hash
        return self

    cdef void _rehash(self):
        self._hash = hash((
            self.start_position, self.end_position, self.peptide_id, self.protein_id,
            self.hypothesis_id, self.glycan_combination_id, self.structure_type,
            self.site_combination_index))

    cpdef glycopeptide_key copy(self, object structure_type=None):
        cdef:
            Py_hash_t _hash
        if structure_type is None:
            structure_type = self.structure_type
        return self.__class__(
            self.start_position, self.end_position, self.peptide_id, self.protein_id,
            self.hypothesis_id, self.glycan_combination_id, structure_type,
            self.site_combination_index)

    def _replace(self, **kwargs):
        cdef glycopeptide_key dup = self.copy()
        for key, val in kwargs.items():
            setattr(dup, key, val)
        dup._rehash()
        return dup

    def __iter__(self):
        yield self.start_position
        yield self.end_position
        yield self.peptide_id
        yield self.protein_id
        yield self.hypothesis_id
        yield self.glycan_combination_id
        yield self.structure_type
        yield self.site_combination_index

    def __reduce__(self):
        return self.__class__, tuple(self)

    def __getitem__(self, i):
        if i == 0:
            return self.start_position
        elif i == 1:
            return self.end_position
        elif i == 2:
            return self.peptide_id
        elif i == 3:
            return self.protein_id
        elif i == 4:
            return self.hypothesis_id
        elif i == 5:
            return self.glycan_combination_id
        elif i == 6:
            return self.structure_type
        elif i == 7:
            return self.site_combination_index
        else:
            raise IndexError(i)

    cpdef dict as_dict(self, bint stringify=False):
        cdef:
            dict d

        d = {}
        if stringify:
            d['start_position'] = str(self.start_position)
            d['end_position'] = str(self.end_position)
            d['peptide_id'] = str(self.peptide_id)
            d['protein_id'] = str(self.protein_id)
            d['hypothesis_id'] = str(self.hypothesis_id)
            d['glycan_combination_id'] = str(self.glycan_combination_id)
            d['structure_type'] = str(self.structure_type)
            d['site_combination_index'] = str(self.site_combination_index)
        else:
            d['start_position'] = self.start_position
            d['end_position'] = self.end_position
            d['peptide_id'] = self.peptide_id
            d['protein_id'] = self.protein_id
            d['hypothesis_id'] = self.hypothesis_id
            d['glycan_combination_id'] = self.glycan_combination_id
            d['structure_type'] = self.structure_type
            d['site_combination_index'] = self.site_combination_index
        return d


@cython.binding(True)
cpdef tuple peptide_backbone_fragment_key(self, target, args, dict kwargs):
    key = ("get_fragments", args, frozenset(kwargs.items()))
    return key


cdef class PeptideDatabaseRecordBase(object):

    def __hash__(self):
        return hash(self.modified_peptide_sequence)

    def __eq__(self, PeptideDatabaseRecordBase other):
        if other is None:
            return False
        if self.id != other.id:
            return False
        if self.protein_id != other.protein_id:
            return False
        if abs(self.calculated_mass - other.calculated_mass) > 1e-3:
            return False
        if self.start_position != other.start_position:
            return False
        if self.end_position != other.end_position:
            return False
        if self.hypothesis_id != other.hypothesis_id:
            return False
        if self.n_glycosylation_sites != other.n_glycosylation_sites:
            return False
        if self.o_glycosylation_sites != other.o_glycosylation_sites:
            return False
        if self.gagylation_sites != other.gagylation_sites:
            return False
        return True

    def __ne__(self, other):
        return not (self == other)

    cpdef bint has_glycosylation_sites(self):
        if len(self.n_glycosylation_sites) > 0:
            return True
        elif len(self.o_glycosylation_sites) > 0:
            return True
        elif len(self.gagylation_sites) > 0:
            return True
        return False

    @classmethod
    def from_record(cls, record):
        return cls(**record)


cdef class GlycopeptideFragmentCachingContext(object):
    def __init__(self, store=None):
        if store is None:
            store = {}
        self.store = store

    cpdef peptide_backbone_fragment_key(self, target, args, dict kwargs):
        key = ("get_fragments", args, frozenset(kwargs.items()))
        return key

    cpdef stub_fragment_key(self, target, args, dict kwargs):
        key = ('stub_fragments', args, frozenset(kwargs.items()), )
        return key

    def __contains__(self, key):
        return PyDict_GetItem(self.store, key) != NULL

    def __getitem__(self, key):
        return self.store[key]

    def __setitem__(self, key, value):
        PyDict_SetItem(self.store, key, value)

    def __reduce__(self):
        return self.__class__, (self.store, )

    def keys(self):
        return self.store.keys()

    def values(self):
        return self.store.values()

    def items(self):
        return self.store.items()

    def clear(self):
        self.store.clear()

    cpdef bind(self, target):
        target.fragment_caches = self
        return target

    cpdef unbind(self, target):
        target.fragment_caches = self.__class__()
        return target

    def __call__(self, target):
        return self.bind(target)

    cpdef _make_target_key(self, tuple key):
        return None


from glycresoft.structure.enums import SpectrumMatchClassification as _StructureClassification

cdef EnumMeta StructureClassification = _StructureClassification


cdef class GlycanAwareGlycopeptideFragmentCachingContext(GlycopeptideFragmentCachingContext):
    cpdef stub_fragment_key(self, target, args, dict kwargs):
        cdef tid = target.id
        key = ('stub_fragments', args, frozenset(
            kwargs.items()), tid.glycan_combination_id, tid.structure_type)
        return key

    cpdef _make_target_key(self, tuple key):
        cdef:
            size_t n, i
            tuple new_key
            PyObject* tmp
            object as_target_peptide
            EnumValue value
        n = PyTuple_GET_SIZE(key)
        value = <EnumValue>PyTuple_GET_ITEM(key, n - 1)
        new_key = PyTuple_New(n)
        for i in range(n - 1):
            tmp = PyTuple_GET_ITEM(key, i)
            Py_INCREF(<object>tmp)
            PyTuple_SET_ITEM(new_key, i, <object>tmp)
        as_target_peptide = StructureClassification[value.value ^ 1]
        Py_INCREF(as_target_peptide)
        PyTuple_SET_ITEM(new_key, n - 1, as_target_peptide)
        return new_key


cpdef list clone_stub_fragments(list stubs):
    cdef:
        size_t i, n
        list result
        StubFragment frag

    n = PyList_GET_SIZE(stubs)
    result = PyList_New(n)
    for i in range(n):
        frag = <StubFragment>PyList_GET_ITEM(stubs, i)
        frag = <StubFragment>StubFragment.clone(frag)
        Py_INCREF(frag)
        PyList_SET_ITEM(result, i, frag)
    return result


@cython.boundscheck(False)
cpdef list clone_and_shift_stub_fragments(list stubs, np.ndarray[float64_t, ndim=1, mode='c'] rand_deltas, bint do_clone=True, int min_shift_size=1):
    cdef:
        size_t i, n, j
        list result
        StubFragment frag
        float64_t delta

    n = PyList_GET_SIZE(stubs)
    if do_clone:
        result = PyList_New(n)
    j = 0
    for i in range(n):
        frag = <StubFragment>PyList_GET_ITEM(stubs, i)
        if do_clone:
            frag = <StubFragment>StubFragment.clone(frag)
        if min_shift_size == 0 or StubFragment.get_glycosylation_size(frag) > min_shift_size:
            delta = rand_deltas[j]
            j += 1
            frag.mass += delta
        if do_clone:
            Py_INCREF(frag)
            PyList_SET_ITEM(result, i, frag)
    if do_clone:
        return result
    return stubs


import array

from typing import (Callable, Collection, DefaultDict,
                    List, TYPE_CHECKING)

import numpy as np

from glycresoft.chromatogram_tree import (
    Unmodified)

from ..spectrum_match import SpectrumMatch, SpectrumSolutionSet

from .base import SolutionEntry, logger, Predicate, always, TargetType, parsimony_sort

if TYPE_CHECKING:
    from .chromatogram import SpectrumMatchSolutionCollectionBase



class RepresenterSelectionStrategy(object):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        pass

    def compute_weights(self, collection: 'SpectrumMatchSolutionCollectionBase', threshold_fn: Callable = always,
                        reject_shifted: bool = False, targets_ignored: Collection = None, require_valid: bool = True) -> List[SolutionEntry]:
        raise NotImplementedError()

    def select(self, representers):
        raise NotImplementedError()

    def sort_solutions(self, representers):
        return parsimony_sort(representers)

    def get_solutions_for_spectrum(self, solution_set: SpectrumSolutionSet, threshold_fn: Callable = always,
                                   reject_shifted: bool = False, targets_ignored: Collection = None, require_valid: bool = True):
        return filter(lambda x: x.valid, solution_set.get_top_solutions(
            d=5,
            reject_shifted=reject_shifted,
            targets_ignored=targets_ignored,
            require_valid=require_valid))

    def __call__(self, collection, threshold_fn: Callable = always, reject_shifted: bool = False,
                 targets_ignored: Collection = None, require_valid: bool = True) -> List[SolutionEntry]:
        return self.compute_weights(
            collection,
            threshold_fn=threshold_fn,
            reject_shifted=reject_shifted,
            targets_ignored=targets_ignored,
            require_valid=require_valid)


class TotalBestRepresenterStrategy(RepresenterSelectionStrategy):
    def compute_weights(self, collection: 'SpectrumMatchSolutionCollectionBase', threshold_fn: Callable = always,
                        reject_shifted: bool = False, targets_ignored: Collection = None, require_valid: bool = True) -> List[SolutionEntry]:
        scores = DefaultDict(float)
        best_scores = DefaultDict(float)
        best_spectrum_match = dict()
        for psm in collection.tandem_solutions:
            if threshold_fn(psm):
                sols = list(self.get_solutions_for_spectrum(
                    psm,
                    reject_shifted=reject_shifted,
                    targets_ignored=targets_ignored,
                    require_valid=require_valid)
                )
                for sol in sols:
                    if not threshold_fn(sol):
                        continue
                    if reject_shifted and sol.mass_shift != Unmodified:
                        continue
                    scores[sol.target] += (sol.score)
                    if best_scores[sol.target] < sol.score:
                        best_scores[sol.target] = sol.score
                        best_spectrum_match[sol.target] = sol

        if len(scores) == 0 and require_valid and collection.tandem_solutions:
            weights = self.compute_weights(
                collection,
                threshold_fn=threshold_fn,
                reject_shifted=reject_shifted,
                targets_ignored=targets_ignored,
                require_valid=False
            )
            if weights:
                logger.warning(
                    f"Failed to find a valid solution for {collection.tandem_solutions}, falling back to previously disqualified solutions")
            return weights

        total = sum(scores.values())
        weights = [
            SolutionEntry(k, v, v / total, best_scores[k],
                          best_spectrum_match[k]) for k, v in scores.items()
            if k in best_spectrum_match
        ]
        weights = self.sort_solutions(weights)
        return weights


class TotalAboveAverageBestRepresenterStrategy(RepresenterSelectionStrategy):
    def compute_weights(self, collection: 'SpectrumMatchSolutionCollectionBase', threshold_fn: Callable = always,
                        reject_shifted: bool = False, targets_ignored: Collection = None, require_valid: bool = True) -> List[SolutionEntry]:
        scores = DefaultDict(lambda: array.array('d'))
        best_scores = DefaultDict(float)
        best_spectrum_match = dict()
        for psm in collection.tandem_solutions:
            if threshold_fn(psm):
                sols = self.get_solutions_for_spectrum(
                    psm,
                    reject_shifted=reject_shifted,
                    targets_ignored=targets_ignored,
                    require_valid=require_valid
                )
                for sol in sols:
                    if not threshold_fn(sol):
                        continue
                    if reject_shifted and sol.mass_shift != Unmodified:
                        continue
                    scores[sol.target].append(sol.score)
                    if best_scores[sol.target] < sol.score:
                        best_scores[sol.target] = sol.score
                        best_spectrum_match[sol.target] = sol

        if len(scores) == 0 and require_valid and collection.tandem_solutions:
            weights = self.compute_weights(
                collection,
                threshold_fn=threshold_fn,
                reject_shifted=reject_shifted,
                targets_ignored=targets_ignored,
                require_valid=False
            )
            if weights:
                logger.warning(
                    f"Failed to find a valid solution for {collection.tandem_solutions}, falling back to previously disqualified solutions")
            return weights

        population = np.concatenate(list(scores.values()))
        min_score = np.mean(population) - np.std(population)
        scores = {k: np.array(v) for k, v in scores.items()}
        thresholded_scores = {k: v[v >= min_score].sum()
                              for k, v in scores.items()}

        total = sum([v for v in thresholded_scores.values()])
        weights = [
            SolutionEntry(k, v, v / total, best_scores[k],
                          best_spectrum_match[k]) for k, v in thresholded_scores.items()
            if k in best_spectrum_match
        ]
        weights = self.sort_solutions(weights)
        return weights



import itertools
from operator import attrgetter
import numpy as np

from glycopeptidepy.structure.glycan import GlycanCompositionProxy

from glycresoft.structure import SpectrumGraph
from glycresoft.task import log_handle

from glycresoft.tandem.oxonium_ions import single_signatures, compound_signatures

from .base import GlycopeptideSpectrumMatcherBase


keyfn = attrgetter("intensity")


class GlycanCompositionSignatureMatcher(GlycopeptideSpectrumMatcherBase):
    minimum_intensity_ratio_threshold = 0.025

    def __init__(self, scan, target, mass_shift=None):
        super(GlycanCompositionSignatureMatcher, self).__init__(scan, target, mass_shift)
        self._init_signature_matcher()

    def _init_signature_matcher(self):
        self.glycan_composition = self._copy_glycan_composition()
        self.expected_matches = dict()
        self.unexpected_matches = dict()
        self.maximum_intensity = float('inf')

    def _copy_glycan_composition(self):
        return GlycanCompositionProxy(self.target.glycan_composition)

    signatures = single_signatures
    compound_signatures = compound_signatures
    all_signatures = single_signatures.copy()
    all_signatures.update(compound_signatures)

    def match(self, error_tolerance=2e-5, rare_signatures=False, *args, **kwargs):
        if len(self.spectrum) == 0:
            return

        gc = self.glycan_composition
        annotations = self.scan.annotations
        if "signature_index_match" in annotations:
            signature_index = annotations['signature_index_match']
            record = signature_index.record_for(gc)
            if record is not None:
                self.maximum_intensity = signature_index.base_peak_intensity
                self.expected_matches = record.expected_matches
                self.unexpected_matches = record.unexpected_matches
                return
            else:
                log_handle.log("No signature ion index entry found for %s in %s" % (gc, self.scan_id))
        self.maximum_intensity = self.base_peak()
        spectrum = self.spectrum

        for mono in self.signatures:
            is_expected = mono.is_expected(self.glycan_composition)
            peak = mono.peak_of(spectrum, error_tolerance)
            if peak is None:
                if is_expected:
                    self.expected_matches[mono] = None
                continue
            if is_expected:
                self.expected_matches[mono] = peak
            else:
                self.unexpected_matches[mono] = peak

        if rare_signatures:
            for compound in self.compound_signatures:
                is_expected = compound.is_expected(self.glycan_composition)
                peak = compound.peak_of(spectrum, error_tolerance)
                if peak is None:
                    if is_expected:
                        self.expected_matches[compound] = None
                    continue
                if is_expected:
                    self.expected_matches[compound] = peak
                else:
                    self.unexpected_matches[compound] = peak

    def _find_peak_pairs(self, error_tolerance=2e-5, include_compound=False, *args, **kwargs):
        peak_set = self.spectrum
        if len(peak_set) == 0:
            return []
        pairs = SpectrumGraph()
        minimum_intensity_threshold = 0.01
        blocks = [(part, part.mass()) for part in self.glycan_composition]
        if include_compound:
            compound_blocks = list(itertools.combinations(self.target, 2))
            compound_blocks = [(block, sum(part.mass() for part in block))
                               for block in compound_blocks]
            blocks.extend(compound_blocks)

        max_peak = self.maximum_intensity
        threshold = max_peak * minimum_intensity_threshold

        for peak in peak_set:
            if peak.intensity < threshold or peak.neutral_mass < 150:
                continue
            for block, mass in blocks:
                for other in peak_set.all_peaks_for(peak.neutral_mass + mass, error_tolerance):
                    if other.intensity < threshold:
                        continue
                    pairs.add(peak, other, block)
        self.spectrum_graph = pairs
        return pairs

    def estimate_missing_ion_importance(self, key) -> float:
        count = key.count_of(self.glycan_composition)
        weight = self.all_signatures[key]
        return min(weight * count, 0.99)

    def _signature_ion_score(self) -> float:
        penalty = 0.0
        for key, peak in self.unexpected_matches.items():
            ratio = peak.intensity / self.maximum_intensity
            if ratio < self.minimum_intensity_ratio_threshold:
                continue
            x = 1 - ratio
            if x <= 0:
                component = 20
            else:
                component = -10 * np.log10(x)
                if np.isnan(component):
                    component = 20
            penalty += component
        for key, peak in self.expected_matches.items():
            matched = peak is not None
            if matched:
                ratio = peak.intensity / self.maximum_intensity
                if ratio < self.minimum_intensity_ratio_threshold:
                    matched = False
            if not matched:
                importance = self.estimate_missing_ion_importance(key)
                x = 1 - importance
                if x <= 0:
                    component = 20
                else:
                    component = -10 * np.log10(x)
                    if np.isnan(component):
                        component = 20
                penalty += component
        return -penalty

    def calculate_score(self, *args, **kwargs):
        self._score = self._signature_ion_score()
        return self._score

    def oxonium_ion_utilization(self, threshold: float=0.05) -> float:
        utilization = 0.0
        for signature, matched in self.expected_matches.items():
            if matched is None or ((matched.intensity / self.maximum_intensity) < threshold):
                comp = 10 * np.log10(1 - self.estimate_missing_ion_importance(signature))
                if np.isnan(comp):
                    comp = -20
                utilization += comp
        return utilization



# -*- coding: utf-8 -*-

'''
Much of this logic is derived from:

    Risk, B. A., Edwards, N. J., & Giddings, M. C. (2013). A peptide-spectrum scoring system
    based on ion alignment, intensity, and pair probabilities. Journal of Proteome Research,
    12(9), 4240–7. http://doi.org/10.1021/pr400286p
'''


import numpy as np
from scipy.special import comb
from decimal import Decimal
import math

from glycopeptidepy.utils.memoize import memoize

from .base import (
    GlycopeptideSpectrumMatcherBase, ChemicalShift, EXDFragmentationStrategy,
    HCDFragmentationStrategy, IonSeries)
from glycresoft.structure import FragmentMatchMap


@memoize(100000000000)
def binomial_pmf(n, i, p):
    try:
        return comb(n, i, exact=True) * (p ** i) * ((1 - p) ** (n - i))
    except OverflowError:
        dn = Decimal(n)
        di = Decimal(i)
        dp = Decimal(p)
        x = math.factorial(dn) / (math.factorial(di) * math.factorial(dn - di))
        return float(x * dp ** di * ((1 - dp) ** (dn - di)))


@memoize(100000000000)
def binomial_tail_probability(n, k, p):
    total = 0.0
    for i in range(k, n):
        v = binomial_pmf(n, i, p)
        if np.isnan(v):
            continue
        total += v
    return total


def binomial_fragments_matched(total_product_ion_count, count_product_ion_matches, ion_tolerance,
                               precursor_mass):
    p = np.exp((np.log(ion_tolerance) + np.log(2)) +
               np.log(count_product_ion_matches) - np.log(precursor_mass))
    return binomial_tail_probability(total_product_ion_count, count_product_ion_matches, p)


def median_sorted(numbers):
    n = len(numbers)
    if n == 0:
        return (n - 1) // 2, 0
    elif n % 2 == 0:
        return (n - 1) // 2, (numbers[(n - 1) // 2] + numbers[((n - 1) // 2) + 1]) / 2.
    else:
        return (n - 1) // 2, numbers[(n - 1) // 2]


def medians(array):
    array.sort()
    offset, m1 = median_sorted(array)
    offset += 1
    i, m2 = median_sorted(array[offset:])
    offset += i + 1
    i, m3 = median_sorted(array[offset:])
    offset += i + 1
    i, m4 = median_sorted(array[offset:])
    return m1, m2, m3, m4


def _counting_tiers(peak_list, matched_peaks, total_product_ion_count):
    intensity_list = np.array([p.intensity for p in peak_list])
    m1, m2, m3, m4 = medians(intensity_list)

    matched_intensities = np.array(
        [p.intensity for p, _ in matched_peaks])
    counts = dict()
    next_count = (matched_intensities > m1).sum()
    counts[1] = next_count

    next_count = (matched_intensities > m2).sum()
    counts[2] = next_count

    next_count = (matched_intensities > m3).sum()
    counts[3] = next_count

    next_count = (matched_intensities > m4).sum()
    counts[4] = next_count
    return counts


def _intensity_tiers(peak_list, matched_peaks, total_product_ion_count):
    intensity_list = np.array([p.intensity for p in peak_list])
    m1, m2, m3, m4 = medians(intensity_list)

    matched_intensities = np.array(
        [p.intensity for p, _ in matched_peaks])
    counts = dict()
    last_count = total_product_ion_count
    next_count = (matched_intensities > m1).sum()
    if last_count == 0:
        counts[1] = 1.0
    elif last_count == next_count:
        counts[1] = binomial_tail_probability(last_count, next_count - 1, 0.5)
    else:
        counts[1] = binomial_tail_probability(last_count, next_count, 0.5)
    last_count = next_count

    next_count = (matched_intensities > m2).sum()
    if last_count == 0:
        counts[2] = 1.0
    elif last_count == next_count:
        counts[2] = binomial_tail_probability(last_count, next_count - 1, 0.5)
    else:
        counts[2] = binomial_tail_probability(last_count, next_count, 0.5)
    last_count = next_count

    next_count = (matched_intensities > m3).sum()
    if last_count == 0:
        counts[3] = 1.0
    elif last_count == next_count:
        counts[3] = binomial_tail_probability(last_count, next_count - 1, 0.5)
    else:
        counts[3] = binomial_tail_probability(last_count, next_count, 0.5)

    last_count = next_count

    next_count = (matched_intensities > m4).sum()
    if last_count == 0:
        counts[4] = 1.0
    elif last_count == next_count:
        counts[4] = binomial_tail_probability(last_count, next_count - 1, 0.5)
    else:
        counts[4] = binomial_tail_probability(last_count, next_count, 0.5)
    return counts


def _score_tiers(peak_list, matched_peaks, total_product_ion_count):
    intensity_list = np.array([p.score for p in peak_list])
    m1, m2, m3, m4 = medians(intensity_list)

    matched_intensities = np.array(
        [p.score for match, p in matched_peaks.items()])
    counts = dict()
    last_count = total_product_ion_count
    next_count = (matched_intensities > m1).sum()
    if last_count == 0:
        counts[1] = 1.0
    elif last_count == next_count:
        counts[1] = binomial_tail_probability(last_count, next_count - 1, 0.5)
    else:
        counts[1] = binomial_tail_probability(last_count, next_count, 0.5)
    last_count = next_count

    next_count = (matched_intensities > m2).sum()
    if last_count == 0:
        counts[2] = 1.0
    elif last_count == next_count:
        counts[2] = binomial_tail_probability(last_count, next_count - 1, 0.5)
    else:
        counts[2] = binomial_tail_probability(last_count, next_count, 0.5)
    last_count = next_count

    next_count = (matched_intensities > m3).sum()
    if last_count == 0:
        counts[3] = 1.0
    elif last_count == next_count:
        counts[3] = binomial_tail_probability(last_count, next_count - 1, 0.5)
    else:
        counts[3] = binomial_tail_probability(last_count, next_count, 0.5)

    last_count = next_count

    next_count = (matched_intensities > m4).sum()
    if last_count == 0:
        counts[4] = 1.0
    elif last_count == next_count:
        counts[4] = binomial_tail_probability(last_count, next_count - 1, 0.5)
    else:
        counts[4] = binomial_tail_probability(last_count, next_count, 0.5)
    return counts


def binomial_intensity(peak_list, matched_peaks, total_product_ion_count):
    if len(matched_peaks) == 0:
        return np.exp(0)
    counts = _intensity_tiers(peak_list, matched_peaks, total_product_ion_count)

    prod = 0
    for k, v in counts.items():
        if v == 0:
            v = 1e-20
        prod += np.log(v)
    return np.exp(prod)


def calculate_precursor_mass(spectrum_match):
    precursor_mass = spectrum_match.target.total_composition().mass
    return precursor_mass


class BinomialSpectrumMatcher(GlycopeptideSpectrumMatcherBase):

    def __init__(self, scan, target, mass_shift=None):
        super(BinomialSpectrumMatcher, self).__init__(scan, target, mass_shift)
        self.solution_map = FragmentMatchMap()
        self._init_binomial()

    def _init_binomial(self):
        self._sanitized_spectrum = set(self.spectrum)
        self.n_theoretical = 0

    def _match_oxonium_ions(self, error_tolerance=2e-5, masked_peaks=None):
        if masked_peaks is None:
            masked_peaks = set()
        val = super(BinomialSpectrumMatcher, self)._match_oxonium_ions(
            error_tolerance=error_tolerance, masked_peaks=masked_peaks)
        self._sanitized_spectrum -= {self.spectrum[i] for i in masked_peaks}
        return val

    def _match_backbone_series(self, series, error_tolerance=2e-5, masked_peaks=None, strategy=None,
                               include_neutral_losses=False):
        if strategy is None:
            strategy = HCDFragmentationStrategy
        for frags in self.get_fragments(series, strategy=strategy, include_neutral_losses=include_neutral_losses):
            # Should this be on the level of position, or the level of the individual fragment ions?
            # At the level of position, this makes missing only glycosylated or unglycosylated ions
            # less punishing, while at the level of the fragment makes more sense by the definition
            # of the geometric mass accuracy interpretation.
            #
            # Using the less severe case to be less pessimistic
            self.n_theoretical += 1
            for frag in frags:
                for peak in self.spectrum.all_peaks_for(frag.mass, error_tolerance):
                    if peak.index.neutral_mass in masked_peaks:
                        continue
                    self.solution_map.add(peak, frag)

    def _sanitize_solution_map(self):
        san = list()
        for pair in self.solution_map:
            if pair.fragment.series != "oxonium_ion":
                san.append(pair)
        return san

    def _compute_average_window_size(self, error_tolerance=2e-5):
        average_window_size = (
            (self.target.peptide_composition(
            ).mass) / 3.) * error_tolerance * 2
        return average_window_size

    def _fragment_matched_binomial(self, error_tolerance=2e-5):
        precursor_mass = calculate_precursor_mass(self)

        fragment_match_component = binomial_fragments_matched(
            self.n_theoretical,
            len(self._sanitize_solution_map()),
            self._compute_average_window_size(error_tolerance),
            precursor_mass
        )
        if fragment_match_component < 1e-170:
            fragment_match_component = 1e-170
        return fragment_match_component

    def _intensity_component_binomial(self):
        intensity_component = binomial_intensity(
            self._sanitized_spectrum,
            self._sanitize_solution_map(),
            self.n_theoretical)

        if intensity_component < 1e-170:
            intensity_component = 1e-170
        return intensity_component

    def _binomial_score(self, error_tolerance=2e-5, *args, **kwargs):
        precursor_mass = calculate_precursor_mass(self)

        solution_map = self._sanitize_solution_map()
        n_matched = len(solution_map)
        if n_matched == 0 or len(self._sanitized_spectrum) == 0:
            return 0

        fragment_match_component = binomial_fragments_matched(
            self.n_theoretical,
            len(solution_map),
            self._compute_average_window_size(error_tolerance),
            precursor_mass
        )

        if fragment_match_component < 1e-170:
            fragment_match_component = 1e-170

        intensity_component = binomial_intensity(
            self._sanitized_spectrum,
            solution_map,
            self.n_theoretical)

        if intensity_component < 1e-170:
            intensity_component = 1e-170
        score = -np.log10(intensity_component) + -np.log10(fragment_match_component)

        if np.isinf(score):
            print("infinite score", self.scan, self.target, intensity_component, fragment_match_component, self.scan)

        return score

    def calculate_score(self, error_tolerance=2e-5, *args, **kwargs):
        score = self._binomial_score(error_tolerance)
        self._score = score
        return score


class StubIgnoringBinomialSpectrumMatcher(BinomialSpectrumMatcher):

    def _sanitize_solution_map(self):
        san = list()
        for pair in self.solution_map:
            if pair.fragment.series not in ("oxonium_ion", "stub_glycopeptide"):
                san.append(pair)
        return san



from glypy.structure.substituent import register as register_substituent_rule
from glypy.composition import formula


substituent_schema = {
    "name": str,
    "composition": str,
    "can_nh_derivatize": bool,
    "is_nh_derivatizable": bool,
    "attachment_composition": str
}


def serialize_substituent_rule(substituent):
    return {
        "name": substituent.name,
        "composition": formula(substituent.composition),
        "can_nh_derivatize": substituent.can_nh_derivatize,
        "is_nh_derivatizable": substituent.is_nh_derivatizable,
        "attachment_composition": formula(substituent.attachment_composition)
    }


def load_substituent_rule(rule_dict):
    return register_substituent_rule(
        rule_dict['name'], rule_dict['composition'],
        rule_dict['can_nh_derivatize'], rule_dict['is_nh_derivatizable'],
        rule_dict['attachment_composition'])


def save_substituent_rule(substituent):
    return serialize_substituent_rule(substituent)



import re
from itertools import product
from collections import deque

from glypy import GlycanComposition as MemoryGlycanComposition
from glypy.structure.glycan_composition import FrozenMonosaccharideResidue
from glypy.io.iupac import IUPACError
from glycresoft.serialize.hypothesis.glycan import GlycanComposition as DBGlycanComposition

from glypy.composition import formula

from .glycan_source import (
    GlycanTransformer, GlycanHypothesisSerializerBase, GlycanCompositionToClass, GlycanTypes, normalize_lookup)
from glycresoft.symbolic_expression import (
    ConstraintExpression, SymbolContext, SymbolNode, ExpressionNode,
    ensuretext)


def descending_combination_counter(counter):
    keys = counter.keys()
    count_ranges = map(lambda lo_hi: range(
        lo_hi[0], lo_hi[1] + 1), counter.values())
    for combination in product(*count_ranges):
        yield dict(zip(keys, combination))


def normalize_iupac_lite(string):
    return str(FrozenMonosaccharideResidue.from_iupac_lite(string))


class CombinatoricCompositionGenerator(object):
    """
    Summary

    Attributes
    ----------
    constraints : list
        Description
    lower_bound : list
        Description
    residue_list : list
        Description
    rules_table : dict
        Description
    upper_bound : list
        Description
    """
    @staticmethod
    def build_rules_table(residue_list, lower_bound, upper_bound):
        rules_table = {}
        for i, residue in enumerate(residue_list):
            lower = lower_bound[i]
            upper = upper_bound[i]
            rules_table[residue] = (lower, upper)
        return rules_table

    def __init__(self, residue_list=None, lower_bound=None, upper_bound=None, constraints=None, rules_table=None,
                 structure_classifiers=None):
        if structure_classifiers is None:
            structure_classifiers = {GlycanTypes.n_glycan: is_n_glycan_classifier}
        self.residue_list = list(map(normalize_iupac_lite, (residue_list or [])))
        self.lower_bound = lower_bound or []
        self.upper_bound = upper_bound or []
        self.constraints = constraints or []
        self.rules_table = rules_table
        self.structure_classifiers = structure_classifiers

        if len(self.constraints) > 0 and not isinstance(self.constraints[0], ConstraintExpression):
            self.constraints = list(
                map(ConstraintExpression.from_list, self.constraints))

        if rules_table is None:
            self._build_rules_table()

        self.normalize_rules_table()
        for constraint in self.constraints:
            self.normalize_symbols(constraint.expression)

        self._iter = None

    def normalize_rules_table(self):
        rules_table = {}
        for key, bounds in self.rules_table.items():
            rules_table[normalize_iupac_lite(key)] = bounds
        self.rules_table = rules_table

    def normalize_symbols(self, expression):
        symbols = list(expression.itersymbols())
        for node in symbols:
            node.symbol = normalize_iupac_lite(node.symbol)

    def _build_rules_table(self):
        rules_table = {}
        for i, residue in enumerate(self.residue_list):
            lower = self.lower_bound[i]
            upper = self.upper_bound[i]
            rules_table[residue] = (lower, upper)
        self.rules_table = rules_table
        return rules_table

    def generate(self):
        for combin in descending_combination_counter(self.rules_table):
            passed = True
            combin = SymbolContext(combin)
            structure_classes = []
            if sum(combin.values()) == 0:
                continue
            for constraint in self.constraints:
                if not constraint(combin):
                    passed = False
                    break
            if passed:
                for name, classifier in self.structure_classifiers.items():
                    if classifier(combin):
                        structure_classes.append(name)
                try:
                    yield MemoryGlycanComposition(**combin.context), structure_classes
                except IUPACError:
                    raise

    __iter__ = generate

    def __next__(self):
        if self._iter is None:
            self._iter = self.generate()
        return next(self._iter)

    next = __next__

    def __repr__(self):
        return "CombinatoricCompositionGenerator\n" + repr(self.rules_table) + '\n' + repr(self.constraints)


class CombinatorialGlycanHypothesisSerializer(GlycanHypothesisSerializerBase):
    def __init__(self, glycan_text_file, database_connection, reduction=None, derivatization=None,
                 hypothesis_name=None, glycan_type=None):
        GlycanHypothesisSerializerBase.__init__(self, database_connection, hypothesis_name)

        self.glycan_file = glycan_text_file
        self.glycan_type = glycan_type
        self.reduction = reduction
        self.derivatization = derivatization

        self.loader = None
        self.transformer = None

    def make_pipeline(self):
        rules, constraints = parse_rules_from_file(self.glycan_file)
        classifiers = None
        if self.glycan_type:
            classifiers = {normalize_lookup(self.glycan_type): lambda x: True}
        self.loader = CombinatoricCompositionGenerator(
            rules_table=rules, constraints=constraints, structure_classifiers=classifiers)
        self.transformer = GlycanTransformer(self.loader, self.reduction, self.derivatization)

    def run(self):
        self.make_pipeline()
        structure_class_lookup = self.structure_class_loader
        acc = []
        self.log("Generating Glycan Compositions from Symbolic Rules for %r" % self.hypothesis)
        counter = 0
        for composition, structure_classes in self.transformer:
            mass = composition.mass()
            composition_string = composition.serialize()
            formula_string = formula(composition.total_composition())
            inst = DBGlycanComposition(
                calculated_mass=mass, formula=formula_string,
                composition=composition_string,
                hypothesis_id=self.hypothesis_id)
            counter += 1
            self.session.add(inst)
            self.session.flush()
            for structure_class in structure_classes:
                structure_class = structure_class_lookup[structure_class]
                acc.append(dict(glycan_id=inst.id, class_id=structure_class.id))
                if len(acc) % 100 == 0:
                    self.session.execute(GlycanCompositionToClass.insert(), acc)
                    acc = []
            if counter % 1000 == 0:
                self.log("%d glycan compositions created" % (counter,))
        if acc:
            self.session.execute(GlycanCompositionToClass.insert(), acc)
            acc = []
        self.session.commit()
        self.log("Generated %d glycan compositions" % counter)


def tryopen(obj):
    if hasattr(obj, "read"):
        return obj
    else:
        return open(obj, 'rt')


def parse_rules_from_file(path):
    """
    Parse monosaccharide bounds and constraints from text file-like
    object

    Parameters
    ----------
    path : file-like or str
        The file to be parsed

    Raises
    ------
    Exception
        Description

    Returns
    -------
    rules_table: dict
        A dict mapping Residue Name to (Lower Bound, Upper Bound)
    constraints: list
        A list of :class:`~.ConstraintExpression` objects
    """
    ranges = []
    constraints = []
    stream = tryopen(path)

    i = 0

    def cast(parts):
        return parts[0], int(parts[1]), int(parts[2])

    for line in stream:
        i += 1
        line = ensuretext(line)

        if line.startswith(";"):
            continue
        parts = re.split(r"\s+", line.replace("\n", ""))
        try:
            if len(parts) == 3:
                ranges.append(cast(parts))
            elif len(parts) == 1:
                if parts[0] in ["\n", "\r", ""]:
                    break
                else:
                    raise Exception("Could not interpret line '%r' at line %d" % (parts, i))
        except Exception as e:
            raise Exception("Failed to interpret line %r at line %d (%r)" % (parts, i, e))

    for line in stream:
        i += 1
        line = ensuretext(line)
        if line.startswith(";"):
            continue
        line = line.replace("\n", "")
        if line in ["\n", "\r", ""]:
            break
        try:
            constraints.append(ConstraintExpression.parse(line))
        except Exception as e:
            raise Exception("Failed to interpret line %r at line %d (%r)" % (line, i, e))

    rules_table = CombinatoricCompositionGenerator.build_rules_table(
        *zip(*ranges))
    try:
        stream.close()
    except Exception:
        pass
    return rules_table, constraints


def write_rules(rules, writer):
    for key, bounds in rules.items():
        lo, hi = bounds
        writer.write("%s %d %d\n" % (key, lo, hi))
    return writer


def write_constraints(constraints, writer):
    writer.write("\n")
    for constraint in constraints:
        writer.write("%s\n" % constraint)
    return writer


class ClassificationConstraint(object):
    """
    Express a classification (label) of a solution by satisfaction
    of a boolean Expression

    Attributes
    ----------
    classification : str
        The label to grant on satisfaction of `constraints`
    constraint : ConstraintExpression
        The constraint that must be satisfied.
    """

    def __init__(self, classification, constraint):
        self.classification = classification
        self.constraint = constraint

    def __call__(self, context):
        """
        Test for satisfaction.

        Parameters
        ----------
        context : SymbolContext

        Returns
        -------
        bool
            Whether the constraint is satisfied
        """
        return self.constraint(context)

    def __repr__(self):
        return "{}\n{}".format(self.classification, self.constraint)


is_n_glycan_classifier = ClassificationConstraint(
    "N-Glycan", (
        ConstraintExpression.from_list(
            ["HexNAc", ">=", "2"]) & ConstraintExpression.from_list(
            ["Hex", ">=", "3"]))
)

is_o_glycan_classifier = ClassificationConstraint(
    "O-Glycan", ((ConstraintExpression.from_list(["HexNAc", ">=", "1"]) &
                  ConstraintExpression.from_list(["Hex", ">=", "1"])) |
                 (ConstraintExpression.from_list(["HexNAc", ">=", "2"])))
)



from glycresoft.tandem import target_decoy
from glycresoft.tandem.target_decoy import (
    TargetDecoyAnalyzer, GroupwiseTargetDecoyAnalyzer)

from .base import SpectrumMatcherBase, PeptideSpectrumMatcherBase
from .intensity_score import LogIntensityScorer
from .simple_score import SimpleCoverageScorer
from .localize import AScoreEvaluator



from collections import defaultdict, OrderedDict

import numpy as np
from scipy import linalg


class GlycanCompositionSolutionRecord(object):
    def __init__(self, glycan_composition, score, total_signal=1.0):
        self.glycan_composition = glycan_composition
        self.score = score
        self.internal_score = self.score
        self.total_signal = total_signal

    def __eq__(self, other):
        if other is None:
            return False
        match = self.glycan_composition == other.glycan_composition
        if not match:
            return match
        match = np.isclose(self.score, other.score)
        if not match:
            return match
        match = np.isclose(self.total_signal, other.total_signal)
        return match

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.glycan_composition)

    @classmethod
    def from_chromatogram(cls, solution):
        return cls(solution.glycan_composition, solution.score,
                   solution.total_signal)

    def __repr__(self):
        return ("{self.__class__.__name__}({self.glycan_composition}, "
                "{self.score}, {self.total_signal})").format(self=self)


def is_diagonal(m):
    if m.shape[0] != m.shape[1]:
        return False
    return (np.count_nonzero(m) - np.count_nonzero(np.diag(m))) == 0


class ObservationWeightState(object):
    def __init__(self, raw_scores, weight_matrix, observed_indices, size):
        self.raw_scores = np.array(raw_scores)
        self.weight_matrix = weight_matrix
        self.observed_indices = observed_indices
        self.size = size
        self.variance_matrix = None
        self.left_inverse_weight_matrix = None
        self.inverse_variance_matrix = None
        self.weighted_scores = None
        if len(self.raw_scores) == 0:
            self.empty()
        else:
            self.transform()

    def empty(self):
        self.variance_matrix = np.array([[]])
        self.left_inverse_weight_matrix = np.array([[]])
        self.inverse_variance_matrix = np.array([[]])
        self.weighted_scores = np.array([])

    def transform(self):
        # This is necessary when the weight matrix is not a square matrix (e.g. the identity matrix)
        # and it is *very slow*. Consider fast-pathing the identity matrix case.
        w = self.weight_matrix.T.dot(self.weight_matrix)
        if is_diagonal(w):
            winv = np.diag([1 / i if i != 0 else i for i in np.diag(w)])
        else:
            winv = linalg.pinv(w)
        self.left_inverse_weight_matrix = winv.dot(self.weight_matrix.T)
        self.variance_matrix = self.left_inverse_weight_matrix.dot(self.left_inverse_weight_matrix.T)
        self.inverse_variance_matrix = w
        self.weighted_scores = self.left_inverse_weight_matrix.dot(self.raw_scores)
        self.weighted_scores = self.weighted_scores[np.nonzero(self.weighted_scores)]

    def expand_variance_matrix(self):
        V = np.zeros((self.size, self.size))
        V[self.observed_indices, self.observed_indices] = np.diag(self.variance_matrix)
        return V


class VariableObservationAggregation(object):
    def __init__(self, network):
        self.aggregation = defaultdict(list)
        self.network = network

    def collect(self, observations):
        for obs in observations:
            self.aggregation[obs.glycan_composition].append(obs)

    def reset(self):
        self.aggregation = defaultdict(list)

    @property
    def total_observations(self):
        q = 0
        for key, values in self.aggregation.items():
            q += len(values)
        return q

    def iterobservations(self):
        for key, values in sorted(self.aggregation.items(), key=lambda x: self.network[x[0]].index):
            for val in values:
                yield val

    def observed_indices(self):
        indices = {self.network[obs.glycan_composition].index for obs in self.iterobservations()}
        return np.array(sorted(indices))

    def calculate_weight(self, observation):
        return 1

    def build_weight_matrix(self):
        q = self.total_observations
        p = len(self.network)
        weights = np.zeros((q, p))
        for i, obs in enumerate(self.iterobservations()):
            weights[i, self.network[obs.glycan_composition].index] = self.calculate_weight(obs)
        return weights

    def estimate_summaries(self):
        E = self.build_weight_matrix()
        scores = [r.score for r in self.iterobservations()]
        return ObservationWeightState(scores, E, self.observed_indices(), len(self.network))

    def build_records(self):
        observation_weights = self.estimate_summaries()
        indices = self.observed_indices()
        nodes = self.network[indices]
        records = []
        indices = []
        for i, node in enumerate(nodes):
            rec = GlycanCompositionSolutionRecord(
                node.glycan_composition, observation_weights.weighted_scores[i],
                sum([rec.total_signal for rec in self.aggregation[node.glycan_composition]]),
            )
            records.append(rec)
            indices.append(node.index)
        return records, observation_weights


class AbundanceWeightedObservationAggregation(VariableObservationAggregation):
    def calculate_weight(self, observation):
        return np.log10(observation.total_signal) / np.log10(1e6)



from .mass_collection import (
    SearchableMassCollection, NeutralMassDatabase)

from .disk_backed_database import (
    GlycopeptideDiskBackedStructureDatabase,
    GlycanCompositionDiskBackedStructureDatabase)



import unittest
import tempfile
import os
import glob

import ms_peak_picker
import ms_deisotope
from ms_deisotope.output import ProcessedMSFileLoader

import numpy as np

from glycresoft.profiler import SampleConsumer

from . import fixtures

agp_glycomics_mzml = fixtures.get_test_data("AGP_Glycomics_20150930_06.centroid.mzML")
agp_glycproteomics_mzml = fixtures.get_test_data("20150710_3um_AGP_001_29_30.mzML")
agp_glycproteomics_mzml_reference = fixtures.get_test_data("20150710_3um_AGP_001_29_30.preprocessed.mzML")


class SampleConsumerBase(object):
    def make_output_directory(self):
        path = tempfile.mkdtemp()
        return path

    def cleanup(self, directory):
        files = glob.glob(os.path.join(directory, "*"))
        for f in files:
            try:
                os.remove(f)
            except OSError:
                pass
        try:
            os.remove(directory)
        except OSError:
            pass


class MSMSSampleConsumerTest(unittest.TestCase, SampleConsumerBase):
    def build_args(self):
        ms1_peak_picking_args = {
            "transforms": [
            ]
        }

        msn_peak_picking_args = {
            "transforms": [
            ]
        }

        ms1_deconvolution_args = {
            "scorer": ms_deisotope.scoring.PenalizedMSDeconVFitter(20.0, 2.0),
            "max_missed_peaks": 3,
            "averagine": ms_deisotope.glycopeptide,
            "truncate_after": SampleConsumer.MS1_ISOTOPIC_PATTERN_WIDTH,
            "ignore_below": SampleConsumer.MS1_IGNORE_BELOW
        }

        msn_deconvolution_args = {
            "scorer": ms_deisotope.scoring.MSDeconVFitter(10.0),
            "averagine": ms_deisotope.peptide,
            "max_missed_peaks": 1,
            "truncate_after": SampleConsumer.MSN_ISOTOPIC_PATTERN_WIDTH,
            "ignore_below": SampleConsumer.MSN_IGNORE_BELOW
        }
        return (
            ms1_peak_picking_args, msn_peak_picking_args,
            ms1_deconvolution_args, msn_deconvolution_args)

    def test_consumer(self):
        (ms1_peak_picking_args, msn_peak_picking_args,
         ms1_deconvolution_args, msn_deconvolution_args) = self.build_args()
        outdir = self.make_output_directory()
        outpath = os.path.join(outdir, "test-output.mzML")

        consumer = SampleConsumer(
            agp_glycproteomics_mzml,
            ms1_peak_picking_args=ms1_peak_picking_args,
            ms1_deconvolution_args=ms1_deconvolution_args,
            msn_peak_picking_args=msn_peak_picking_args,
            msn_deconvolution_args=msn_deconvolution_args,
            storage_path=outpath, sample_name=None,
            n_processes=5,
            extract_only_tandem_envelopes=True,
            ms1_averaging=1)
        consumer.start()

        reader = ProcessedMSFileLoader(outpath)
        reference = ProcessedMSFileLoader(agp_glycproteomics_mzml_reference)

        for a_bunch, b_bunch in zip(reader, reference):
            assert a_bunch.precursor.id == b_bunch.precursor.id
            assert len(a_bunch.products) == len(b_bunch.products)
            for a_product, b_product in zip(a_bunch.products, b_bunch.products):
                assert a_product.precursor_information.defaulted == b_product.precursor_information.defaulted
                matched = np.isclose(a_product.precursor_information.neutral_mass,
                                     b_product.precursor_information.neutral_mass)
                message = ["%0.3f not close to %0.3f for %s of %s" % (
                    a_product.precursor_information.neutral_mass,
                    b_product.precursor_information.neutral_mass,
                    a_product.id, a_product.precursor_information.precursor_scan_id)]
                message.append("Found precursor score %r, expected %r" % (
                    a_product.precursor_information.precursor.deconvoluted_peak_set.has_peak(
                        a_product.precursor_information.neutral_mass).score,
                    b_product.precursor_information.precursor.deconvoluted_peak_set.has_peak(
                        b_product.precursor_information.neutral_mass).score
                ))
                assert matched, '\n'.join(message)
                assert len(a_product.deconvoluted_peak_set) == len(b_product.deconvoluted_peak_set)

        reader.close()
        reference.close()

        self.cleanup(outdir)


@unittest.skip("This test is subsumed by MSMSSampleConsumerTest")
class SampleConsumerTest(unittest.TestCase, SampleConsumerBase):
    def build_args(self):
        ms1_peak_picking_args = {
            "transforms": [
            ]
        }

        msn_peak_picking_args = {
            "transforms": [
            ]
        }

        ms1_deconvolution_args = {
            "scorer": ms_deisotope.scoring.PenalizedMSDeconVFitter(35.0, 2.0),
            "max_missed_peaks": 1,
            "averagine": ms_deisotope.glycan,
            "truncate_after": SampleConsumer.MS1_ISOTOPIC_PATTERN_WIDTH,
            "ignore_below": SampleConsumer.MS1_IGNORE_BELOW
        }

        msn_deconvolution_args = {
            "scorer": ms_deisotope.scoring.MSDeconVFitter(10.0),
            "averagine": ms_deisotope.glycan,
            "max_missed_peaks": 1,
            "truncate_after": SampleConsumer.MSN_ISOTOPIC_PATTERN_WIDTH,
            "ignore_below": SampleConsumer.MSN_IGNORE_BELOW
        }
        return (
            ms1_peak_picking_args, msn_peak_picking_args,
            ms1_deconvolution_args, msn_deconvolution_args)

    def test_consumer(self):
        (ms1_peak_picking_args, msn_peak_picking_args,
         ms1_deconvolution_args, msn_deconvolution_args) = self.build_args()
        outdir = self.make_output_directory()
        outpath = os.path.join(outdir, "test-output.mzML")

        consumer = SampleConsumer(
            agp_glycomics_mzml,
            ms1_peak_picking_args=ms1_peak_picking_args,
            ms1_deconvolution_args=ms1_deconvolution_args,
            msn_peak_picking_args=msn_peak_picking_args,
            msn_deconvolution_args=msn_deconvolution_args,
            storage_path=outpath, sample_name=None,
            n_processes=5,
            extract_only_tandem_envelopes=False)
        consumer.start()

        reader = ProcessedMSFileLoader(outpath)

        scan = reader.get_scan_by_id("scanId=1601016")
        self.assertIsNotNone(scan.deconvoluted_peak_set.has_peak(958.66, use_mz=1))

        reader.close()

        self.cleanup(outdir)


if __name__ == '__main__':
    unittest.main()



from . import mass_shift
from .mass_shift import (
    MassShift, CompoundMassShift, Unmodified, Ammonium,
    Sodium, Formate, Potassium)

from . import chromatogram
from .chromatogram import (
    Chromatogram, ChromatogramInterface, ChromatogramTreeNode,
    ChromatogramTreeList, EmptyListException, DuplicateNodeError,
    mask_subsequence, split_by_charge, group_by, ChromatogramWrapper,
    GlycanCompositionChromatogram, GlycopeptideChromatogram, get_chromatogram)


from . import grouping
from .grouping import (
    ChromatogramForest, ChromatogramOverlapSmoother,
    smooth_overlaps, build_rt_interval_tree)


from . import generic
from .generic import (
    SimpleChromatogram,
    SimpleEntityChromatogram,
    find_truncation_points)

from . import index
from .index import ChromatogramFilter, DisjointChromatogramSet


from . import relation_graph
from .relation_graph import ChromatogramGraph


from . import mass_shift_tree
from .mass_shift_tree import (prune_bad_mass_shift_branches, MassShiftTreePruner)


from . import utils
from .utils import ArithmeticMapping



import os
import functools

from collections import defaultdict

import click

import ms_deisotope
import glypy

from ms_deisotope import MSFileLoader
from ms_deisotope.output import ProcessedMSFileLoader
from ms_deisotope.feature_map import quick_index
from ms_deisotope.tools.deisotoper import deisotope

from glycresoft.cli.base import cli

from glycresoft.profiler import SampleConsumer


@cli.group('mzml', short_help='Inspect and preprocess mzML files')
def mzml_cli():
    pass


preprocess = click.Command(
    "preprocess",
    context_settings=deisotope.context_settings,
    callback=functools.partial(deisotope.callback, workflow_cls=SampleConsumer, configure_logging=False),
    params=deisotope.params,
    help=deisotope.help,
    epilog=deisotope.epilog,
)

mzml_cli.add_command(preprocess, "preprocess")


@mzml_cli.command('rt-to-id', short_help="Look up the retention time for a given scan id")
@click.argument("ms-file", type=click.Path(exists=True))
@click.argument("rt", type=float)
def rt_to_id(ms_file, rt):
    loader = MSFileLoader(ms_file)
    id = loader._locate_ms1_scan(loader.get_scan_by_time(rt)).id
    click.echo(id)


@mzml_cli.command("info", short_help='Summary information describing a processed mzML file')
@click.argument("ms-file", type=click.Path(exists=True, file_okay=True, dir_okay=False))
def msfile_info(ms_file):
    reader = ProcessedMSFileLoader(ms_file)
    if not reader.has_index_file():
        index, intervals = quick_index.index(ms_deisotope.MSFileLoader(ms_file))
        reader.extended_index = index
        with open(reader._index_file_name, 'w') as handle:
            index.serialize(handle)
    click.echo("Name: %s" % (os.path.basename(ms_file),))
    click.echo("MS1 Scans: %d" % (len(reader.extended_index.ms1_ids),))
    click.echo("MSn Scans: %d" % (len(reader.extended_index.msn_ids),))

    n_defaulted = 0
    n_orphan = 0

    charges = defaultdict(int)
    first_msn = float('inf')
    last_msn = 0
    for scan_info in reader.extended_index.msn_ids.values():
        n_defaulted += scan_info.get('defaulted', False)
        n_orphan += scan_info.get('orphan', False)
        charges[scan_info['charge']] += 1
        rt = scan_info['scan_time']
        if rt < first_msn:
            first_msn = rt
        if rt > last_msn:
            last_msn = rt

    click.echo("First MSn Scan: %0.2f Minutes" % (first_msn,))
    click.echo("Last MSn Scan: %0.2f Minutes" % (last_msn,))

    for charge, count in sorted(charges.items()):
        if not isinstance(charge, int):
            continue
        click.echo("Precursors with Charge State %d: %d" % (charge, count))

    click.echo("Defaulted MSn Scans: %d" % (n_defaulted,))
    click.echo("Orphan MSn Scans: %d" % (n_orphan,))


@mzml_cli.command("oxonium-signature", short_help=(
    'Report Oxonium Ion Signature Score (G-Score) and Glycan Structure Signature Score'
    ' for each MSn scan in a processed mzML file'))
@click.argument("ms-file", type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option("-g", "--g-score-threshold", type=float, default=0.05, help="Minimum G-Score to report")
def oxonium_signature(ms_file, g_score_threshold=0.05):
    reader = ProcessedMSFileLoader(ms_file)
    if not reader.has_index_file():
        click.secho("Building temporary index...", fg='yellow')
        index, intervals = quick_index.index(ms_deisotope.MSFileLoader(ms_file))
        reader.extended_index = index
        with open(reader._index_file_name, 'w') as handle:
            index.serialize(handle)

    from glycresoft.tandem.glycan.scoring.signature_ion_scoring import SignatureIonScorer
    from glycresoft.tandem.oxonium_ions import gscore_scanner

    refcomp = glypy.GlycanComposition.parse("{Fuc:1; Hex:5; HexNAc:4; Neu5Ac:1; NeuGc:1}")
    headers = ["scan_id", "precursor_mass", "precursor_charge", "g_score", "signature_score"]
    click.echo("\t".join(headers))
    for scan_id in reader.extended_index.msn_ids.keys():
        scan = reader.get_scan_by_id(scan_id)
        gscore = gscore_scanner(scan.deconvoluted_peak_set)
        if gscore >= g_score_threshold:
            signature_match = SignatureIonScorer.evaluate(
                scan, refcomp, use_oxonium=False)
            click.echo("%s\t%f\t%r\t%f\t%f" % (
                scan_id, scan.precursor_information.neutral_mass,
                scan.precursor_information.charge, gscore,
                signature_match.score))



cimport cython

from libc.stdlib cimport malloc, free

from cpython cimport (
    PyTuple_GetItem, PyTuple_GET_ITEM, PyTuple_GET_SIZE,
    PyInt_AsLong, PyFloat_AsDouble,
    PyObject, PyDict_GetItem, PyDict_SetItem,
    PyList_GET_SIZE, PyList_GET_ITEM, PyDict_Next)

from glypy._c.structure.glycan_composition cimport _CompositionBase
from ms_deisotope._c.peak_set cimport DeconvolutedPeak, DeconvolutedPeakSet
from ms_deisotope._c.utils cimport _peak_sequence_bp

from glycopeptidepy._c.structure.fragment cimport SimpleFragment

from glycresoft._c.structure.fragment_match_map cimport PeakFragmentPair

import warnings

from collections import defaultdict

import glypy
from glycopeptidepy import PeptideSequence
from glypy.structure.glycan_composition import FrozenMonosaccharideResidue
from glycopeptidepy.structure.glycan import GlycanCompositionProxy as _GlycanCompositionProxy

cdef object from_iupac_lite = FrozenMonosaccharideResidue.from_iupac_lite
cdef object warn = warnings.warn
cdef object GlycanCompositionProxy = _GlycanCompositionProxy

cdef class SignatureSpecification(object):
    def __init__(self, components, masses):
        self.components = tuple(FrozenMonosaccharideResidue.from_iupac_lite(k) for k in components)
        self.masses = tuple(masses)
        self._hash = hash(self.components)
        self._is_compound = len(self.masses) > 1
        self.n_masses = PyTuple_GET_SIZE(self.masses)
        self._init_mass_array()

    cdef void _init_mass_array(self):
        self._masses = <double*>malloc(sizeof(double) * self.n_masses)
        for i in range(self.n_masses):
            obj = self.masses[i]
            self._masses[i] = PyFloat_AsDouble(obj)

    def __dealloc__(self):
        if self._masses != NULL:
            free(self._masses)
            self._masses = NULL

    def __getitem__(self, i):
        return <object>PyTuple_GetItem(self.components, i)

    def __iter__(self):
        return iter(self.components)

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if other is None:
            return False
        if isinstance(other, SignatureSpecification):
            return self.components == (<SignatureSpecification>other).components
        else:
            return self.components == other.components

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "{self.__class__.__name__}({self.components}, {self.masses})".format(self=self)

    cpdef bint is_expected(self, glycan_composition):
        cdef:
            size_t i, n
            bint is_expected
            long count
            _CompositionBase composition

        if isinstance(glycan_composition, GlycanCompositionProxy):
            glycan_composition = glycan_composition.obj
            while isinstance(glycan_composition, GlycanCompositionProxy):
                glycan_composition = glycan_composition.obj
            composition = <_CompositionBase?>glycan_composition
        elif isinstance(glycan_composition, _CompositionBase):
            composition = <_CompositionBase?>glycan_composition
        else:
            raise TypeError("Requires a _CompositionBase or GlycanCompositionProxy")

        n = PyTuple_GET_SIZE(self.components)
        for i in range(n):
            component = <object>PyTuple_GET_ITEM(self.components, i)
            tmp = composition._getitem_fast(component)
            count = PyInt_AsLong(tmp)
            if count == 0:
                return False
        return True

    cpdef int count_of(self, glycan_composition):
        cdef:
            size_t i, n
            bint is_expected
            long count, limit
            _CompositionBase composition

        if isinstance(glycan_composition, GlycanCompositionProxy):
            glycan_composition = glycan_composition.obj
            while isinstance(glycan_composition, GlycanCompositionProxy):
                glycan_composition = glycan_composition.obj
            composition = <_CompositionBase?>glycan_composition
        elif isinstance(glycan_composition, _CompositionBase):
            composition = <_CompositionBase?>glycan_composition
        else:
            raise TypeError("Requires a _CompositionBase or GlycanCompositionProxy")


        n = PyTuple_GET_SIZE(self.components)
        limit = 100000
        for i in range(n):
            component = <object>PyTuple_GET_ITEM(self.components, i)
            count = PyInt_AsLong(composition._getitem_fast(component))
            if count < limit:
                limit = count
        return limit

    cpdef DeconvolutedPeak peak_of(self, DeconvolutedPeakSet spectrum, double error_tolerance):
        cdef:
            size_t i, j, n
            double mass, best_signal
            DeconvolutedPeak peak, next_peak
            tuple peaks

        peak = None
        best_signal = -1
        for j in range(self.n_masses):
            mass = self._masses[j]
            peaks = spectrum.all_peaks_for(mass, error_tolerance)
            n = PyTuple_GET_SIZE(peaks)
            for i in range(n):
                next_peak = <DeconvolutedPeak>PyTuple_GET_ITEM(peaks, i)
                if next_peak.intensity > best_signal:
                    peak = next_peak
                    best_signal = next_peak.intensity
        return peak


@cython.freelist(5000)
cdef class SignatureIonMatchRecord:

    @staticmethod
    cdef SignatureIonMatchRecord _create(dict expected_matches, dict unexpected_matches, object glycan_composition):
        cdef SignatureIonMatchRecord self = SignatureIonMatchRecord.__new__(SignatureIonMatchRecord)
        self.expected_matches = expected_matches
        self.unexpected_matches = unexpected_matches
        self.glycan_composition = glycan_composition
        self.score = 0
        return self

    def __init__(self, dict expected_matches, dict unexpected_matches, object glycan_composition, double score=0):
        self.expected_matches = expected_matches
        self.unexpected_matches = unexpected_matches
        self.glycan_composition = glycan_composition
        self.score = score

    cpdef match(self, list signatures, DeconvolutedPeakSet spectrum, double error_tolerance=2e-5):
        cdef:
            size_t i, n
            SignatureSpecification specification
            bint is_expected
            DeconvolutedPeak peak

        n = PyList_GET_SIZE(signatures)
        for i in range(n):
            specification = <SignatureSpecification>PyList_GET_ITEM(signatures, i)
            is_expected = specification.is_expected(self.glycan_composition)
            peak = specification.peak_of(spectrum, error_tolerance)

            if peak is None:
                if is_expected:
                    PyDict_SetItem(self.expected_matches, specification, None)
                continue
            if is_expected:
                PyDict_SetItem(self.expected_matches, specification, peak)
            else:
                PyDict_SetItem(self.unexpected_matches, specification, peak)

    def __repr__(self):
        template = "{self.__class__.__name__}({self.expected_matches}, {self.unexpected_matches}, {self.glycan_composition})"
        return template.format(self=self)


@cython.freelist(2000)
cdef class SignatureIonIndexMatch:
    @staticmethod
    cdef SignatureIonIndexMatch _create(dict glycan_to_key, dict key_to_record, double base_peak_intensity):
        cdef SignatureIonIndexMatch self = SignatureIonIndexMatch.__new__(SignatureIonIndexMatch)
        self.glycan_to_key = glycan_to_key
        self.key_to_record = key_to_record
        self.base_peak_intensity = base_peak_intensity
        return self

    cpdef SignatureIonMatchRecord record_for(self, object glycan_composition):
        cdef:
            object key
            PyObject* tmp
        key = str(glycan_composition)
        tmp = PyDict_GetItem(self.glycan_to_key, key)
        if tmp == NULL:
            warnings.warn("{} not found in signature ion index".format(key))
            return None
        key = <object>tmp
        tmp = PyDict_GetItem(self.key_to_record, key)
        if tmp == NULL:
            return None
        return <SignatureIonMatchRecord>tmp

    def __repr__(self):
        template = f"{self.__class__.__name__}(<{len(self.key_to_record)}>)"
        return template


cdef class SignatureIonIndex:

    def __init__(self, signatures):
        self.signatures = signatures
        self.glycan_to_key = {}
        self.key_to_representative = {}

    def build_index(self, glycan_composition_records):
        index = defaultdict(list)

        for gc_rec in glycan_composition_records:
            key = tuple(sig.count_of(gc_rec.composition) for sig in self.signatures)
            index[key].append(gc_rec.composition)

        key_to_representative = {}
        reverse_index = {}
        for key, gcs in index.items():
            for gc in gcs:
                reverse_index[str(gc)] = key
            key_to_representative[key] = gcs[0]
        self.glycan_to_key = reverse_index
        self.key_to_representative = key_to_representative

    cpdef SignatureIonIndexMatch match(self, DeconvolutedPeakSet spectrum, double error_tolerance=2e-5):
        cdef:
            PyObject* pkey
            PyObject* pval
            Py_ssize_t pos
            dict result
            object key, value
            SignatureIonMatchRecord record
            DeconvolutedPeak peak
            double base_peak_intensity
        result = {}
        base_peak_intensity = 0.0
        pos = 0
        while PyDict_Next(self.key_to_representative, &pos, &pkey, &pval):
            if pkey != NULL:
                key = <object>pkey
                value = <object>pval
                record = SignatureIonMatchRecord._create({}, {}, value)
                record.match(self.signatures, spectrum, error_tolerance)
                PyDict_SetItem(result, key, record)

        peak = _peak_sequence_bp(None, spectrum)
        if peak is not None:
            base_peak_intensity = peak.intensity
        return SignatureIonIndexMatch._create(self.glycan_to_key, result, base_peak_intensity)


@cython.freelist(1000)
cdef class OxoniumIndexMatch(object):

    def __init__(self, index_matches, glycan_to_index, id_to_index):
        self.index_matches = index_matches
        self.glycan_to_index = glycan_to_index
        self.id_to_index = id_to_index

    @staticmethod
    cdef OxoniumIndexMatch _create(dict index_matches, dict glycan_to_index, dict id_to_index):
        cdef OxoniumIndexMatch self = OxoniumIndexMatch.__new__(OxoniumIndexMatch)
        self.index_matches = index_matches
        self.glycan_to_index = glycan_to_index
        self.id_to_index = id_to_index
        return self

    cpdef list by_glycan(self, glycan):
        cdef:
            PyObject* tmp

        tmp = PyDict_GetItem(self.glycan_to_index, glycan)
        if tmp == NULL:
            return None
        index = <object>tmp
        tmp = PyDict_GetItem(self.index_matches, index)
        if tmp == NULL:
            return None
        return <list>tmp

    cpdef list by_id(self, glycan_id):
        cdef:
            PyObject* tmp

        tmp = PyDict_GetItem(self.id_to_index, glycan_id)
        if tmp == NULL:
            return None
        index = <object>tmp
        tmp = PyDict_GetItem(self.index_matches, index)
        if tmp == NULL:
            return None
        return <list>tmp


cdef class OxoniumIndex(object):
    '''An index for quickly matching all oxonium ions against a spectrum and efficiently mapping them
    back to individual glycan compositions.
    '''

    def __init__(self, fragments=None, fragment_index=None, glycan_to_index=None, index_to_simplified_index=None):
        self.fragments = fragments or []
        self.fragment_index = fragment_index or {}
        self.glycan_to_index = glycan_to_index or {}
        self.index_to_glycan = self.invert_glycan_index()
        self.index_to_simplified_index = index_to_simplified_index or {}

    def invert_glycan_index(self):
        cdef:
            PyObject* tmp
            PyObject* pkey
            PyObject* pval
            Py_ssize_t pos
            list acc
            list id_bucket
            dict index_to_glycan

        pos = 0
        index_to_glycan = {}
        while PyDict_Next(self.glycan_to_index, &pos, &pkey, &pval):
            id_bucket = <list>pval
            key = <object>pkey
            for val in id_bucket:
                tmp = PyDict_GetItem(index_to_glycan, val)
                if tmp == NULL:
                    acc = []
                    PyDict_SetItem(index_to_glycan, val, acc)
                else:
                    acc = <list>tmp
                acc.append(key)
        return index_to_glycan

    def _make_glycopeptide_stub(self, glycan_composition):
        p = PeptideSequence("P%s" % glycan_composition)
        return p

    def build_index(self, glycan_composition_records, **kwargs):
        cdef:
            PyObject* tmp
            list acc
            list id_bucket
            dict fragments, fragment_index, glycan_index

        fragments = {}
        fragment_index = {}
        glycan_index = {}
        for gc_rec in glycan_composition_records:
            tmp = PyDict_GetItem(glycan_index, gc_rec.composition)
            if tmp == NULL:
                id_bucket = []
                PyDict_SetItem(glycan_index, gc_rec.composition, id_bucket)
            else:
                id_bucket = <list>tmp
            id_bucket.append(gc_rec.id)

            p = self._make_glycopeptide_stub(gc_rec.composition)
            for frag in p.glycan_fragments(**kwargs):
                fragments[frag.name] = frag
                tmp = PyDict_GetItem(fragment_index, frag.name)
                if tmp == NULL:
                    acc = []
                    PyDict_SetItem(fragment_index, frag.name, acc)
                else:
                    acc = <list>tmp
                acc.append(gc_rec.id)

        self.glycan_to_index = glycan_index
        self.fragment_index = fragment_index
        self.fragments = sorted(fragments.values(), key=lambda x: x.mass)
        self.index_to_glycan = self.invert_glycan_index()
        self.simplify()

    cpdef OxoniumIndexMatch match(self, DeconvolutedPeakSet spectrum, double error_tolerance):
        cdef:
            dict match_index
            size_t i, n, j, m
            DeconvolutedPeak peak
            SimpleFragment fragment
            list bucket, acc
            PyObject* tmp
        match_index = {}
        n = PyList_GET_SIZE(self.fragments)
        for i in range(n):
            fragment = <SimpleFragment>PyList_GET_ITEM(self.fragments, i)
            peak = spectrum.has_peak(fragment.mass, error_tolerance)
            if peak is not None:
                bucket = <list>PyDict_GetItem(self.fragment_index, fragment.name)
                m = PyList_GET_SIZE(bucket)
                for j in range(m):
                    key = <object>PyList_GET_ITEM(bucket, j)
                    tmp = PyDict_GetItem(match_index, key)
                    if tmp == NULL:
                        acc = []
                        PyDict_SetItem(match_index, key, acc)
                    else:
                        acc = <list>tmp
                    acc.append(PeakFragmentPair._create_simple(peak, fragment))
        return OxoniumIndexMatch._create(match_index, self.glycan_to_index, self.index_to_simplified_index)

    cpdef object simplify(self):
        cdef:
            PyObject* tmp
            list acc
        id_to_frag_group = defaultdict(set)
        for f, members in self.fragment_index.items():
            for member in members:
                id_to_frag_group[member].add(f)

        groups = defaultdict(list)
        for member, group in id_to_frag_group.items():
            groups[frozenset(group)].append(member)

        counter = 0
        new_fragment_index = {}
        new_glycan_index = {}
        secondary_index = {}
        for frag_group, members in groups.items():
            new_id = counter
            counter += 1
            for frag in frag_group:
                tmp = PyDict_GetItem(new_fragment_index, frag)
                if tmp == NULL:
                    acc = []
                    PyDict_SetItem(new_fragment_index, frag, acc)
                else:
                    acc = <list>tmp
                acc.append(new_id)

            for member in members:
                for key_for_member in self.index_to_glycan[member]:
                    new_glycan_index[key_for_member] = new_id
                secondary_index[member] = new_id

        self.index_to_simplified_index = secondary_index
        self.glycan_to_index = new_glycan_index
        self.fragment_index = new_fragment_index
        self.index_to_glycan = {}
        for k, v in self.glycan_to_index.items():
            tmp = PyDict_GetItem(self.index_to_glycan, v)
            if tmp == NULL:
                acc = []
                PyDict_SetItem(self.index_to_glycan, v, acc)
            else:
                acc = <list>tmp
            acc.append(k)



from collections import namedtuple

import numpy as np
try:
    from matplotlib import pyplot as plt
except ImportError:
    pass

from scipy import stats

from .linear_regression import SMALL_ERROR


CalibrationPoint = namedtuple("CalibrationPoint", (
    "reference_index", "reference_point_rt", "reference_point_rt_pred",
    # rrt = Relative Retention Time, prrt = Predicted Relative Retention Time
    "rrt", "prrt", "residuals", 'weight'))


class RecalibratingPredictor(object):
    def __init__(self, training_examples, testing_examples, model, scale=1.0, dilation=1.0, weighted=True):
        if training_examples is None:
            training_examples = []
        self.training_examples = np.array(training_examples)
        self.testing_examples = np.array(testing_examples)
        self.model = model
        self.scale = scale
        self.dilation = dilation
        self.configurations = dict()
        self._fit()
        self.apex_time_array = np.array(
            [self.model._get_apex_time(c) for c in testing_examples])
        if weighted:
            self.weight_array = np.array(
                [np.log10(c.total_signal) for c in testing_examples])
            self.weight_array /= self.weight_array.max()
        else:
            self.weight_array = np.ones_like(self.apex_time_array)
        self.weighted = weighted
        self.weight_array /= self.weight_array.max()
        self.predicted_apex_time_array = self._predict()
        self.score_array = self._score()

    def _adapt_dilate_fit(self, reference_point, dilation):
        parameters = np.hstack(
            [self.model.parameters[0], dilation * self.model.parameters[1:]])
        predicted_reference_point_rt = self.model._prepare_data_vector(
            reference_point).dot(parameters)
        reference_point_rt = self.model._get_apex_time(reference_point)
        resid = []
        relative_retention_time = []
        predicted_relative_retention_time = []
        for ex in self.testing_examples:
            rrt = self.model._get_apex_time(ex) - reference_point_rt
            prrt = self.model._prepare_data_vector(ex).dot(
                parameters) - predicted_reference_point_rt
            relative_retention_time.append(rrt)
            predicted_relative_retention_time.append(prrt)
            resid.append(prrt - rrt)
        return (reference_point_rt, predicted_reference_point_rt,
                relative_retention_time, predicted_relative_retention_time, resid)

    def _predict_delta_single(self, test_point, reference_point, dilation=None):
        if dilation is None:
            dilation = self.dilation
        parameters = np.hstack(
            [self.model.parameters[0], dilation * self.model.parameters[1:]])
        predicted_reference_point_rt = self.model._prepare_data_vector(
            reference_point).dot(parameters)
        reference_point_rt = self.model._get_apex_time(reference_point)
        rrt = self.model._get_apex_time(test_point) - reference_point_rt
        prrt = self.model._prepare_data_vector(test_point).dot(
            parameters) - predicted_reference_point_rt
        return prrt - rrt

    def predict_delta_single(self, test_point, dilation=None):
        if dilation is None:
            dilation = self.dilation
        delta = []
        weight = []
        for _i, reference_point in enumerate(self.testing_examples):
            delta.append(self._predict_delta_single(
                test_point, reference_point, dilation))
            weight.append(np.log10(reference_point.total_signal))
        return np.dot(delta, weight) / np.sum(weight), np.std(delta)

    def score_single(self, test_point):
        delta, _sd = (self.predict_delta_single(test_point))
        delta = abs(delta)
        score = stats.t.sf(
            delta,
            df=self._df(), scale=self.scale) * 2
        score -= SMALL_ERROR
        if score < SMALL_ERROR:
            score = SMALL_ERROR
        return score

    def _fit(self):
        for i, training_reference_point in enumerate(self.training_examples):
            reference_point = [
                c for c in self.testing_examples
                if (c.glycan_composition == training_reference_point.glycan_composition)]
            if not reference_point:
                continue
            reference_point = max(
                reference_point, key=lambda x: x.total_signal)
            dilation = self.dilation
            (reference_point_rt, predicted_reference_point_rt,
             relative_retention_time,
             predicted_relative_retention_time, resid) = self._adapt_dilate_fit(reference_point, dilation)
            self.configurations[i] = CalibrationPoint(
                i, reference_point_rt, predicted_reference_point_rt,
                np.array(relative_retention_time), np.array(
                    predicted_relative_retention_time),
                np.array(resid), np.log10(reference_point.total_signal)
            )
        if not self.configurations:
            for i, reference_point in enumerate(self.testing_examples):
                dilation = self.dilation
                (reference_point_rt, predicted_reference_point_rt,
                 relative_retention_time,
                 predicted_relative_retention_time, resid) = self._adapt_dilate_fit(reference_point, dilation)
                self.configurations[-i] = CalibrationPoint(
                    i, reference_point_rt, predicted_reference_point_rt,
                    np.array(relative_retention_time), np.array(
                        predicted_relative_retention_time),
                    np.array(resid), np.log10(reference_point.total_signal)
                )

    def _predict(self):
        configs = self.configurations
        predicted_apex_time_array = []
        weight = []
        for _key, calibration_point in configs.items():
            predicted_apex_time_array.append(
                (calibration_point.prrt + calibration_point.reference_point_rt) * calibration_point.weight)
            weight.append(calibration_point.weight)
        return np.sum(predicted_apex_time_array, axis=0) / np.sum(weight)

    def _df(self):
        return max(len(self.configurations) - len(self.model.parameters), 1)

    def _score(self):
        score = stats.t.sf(
            abs(self.predicted_apex_time_array - self.apex_time_array),
            df=self._df(), scale=self.scale) * 2
        score -= SMALL_ERROR
        score[score < SMALL_ERROR] = SMALL_ERROR
        return score

    def R2(self, adjust=False):
        y = self.apex_time_array
        w = self.weight_array
        yhat = self.predicted_apex_time_array
        residuals = (y - yhat)
        rss = (w * residuals * residuals).sum()
        tss = (y - y.mean())
        tss = (w * tss * tss).sum()
        if adjust:
            n = len(y)
            k = len(self.model.parameters)
            adjustment_factor = (n - 1.0) / float(n - k - 1.0)
        else:
            adjustment_factor = 1.0
        R2 = (1 - adjustment_factor * (rss / tss))
        return R2

    @classmethod
    def predict(cls, training_examples, testing_examples, model, dilation=1.0):
        inst = cls(training_examples, testing_examples, model, dilation)
        return inst.predicted_apex_time_array

    @classmethod
    def score(cls, training_examples, testing_examples, model, dilation=1.0):
        inst = cls(training_examples, testing_examples, model, dilation)
        return inst.score_array

    def plot(self, ax=None):
        if ax is None:
            _fig, ax = plt.subplots(1)
        X = self.apex_time_array
        Y = self.predicted_apex_time_array
        S = np.array([c.total_signal for c in self.testing_examples])
        S /= S.max()
        S *= 100
        ax.scatter(X, Y, s=S, marker='o')
        ax.plot((X.min(), X.max()), (X.min(), X.max()),
                color='black', linestyle='--', lw=0.75)
        ax.set_xlabel('Experimental Apex Time', fontsize=18)
        ax.set_ylabel('Predicted Apex Time', fontsize=18)
        ax.figure.text(0.8, 0.15, "$R^2=%0.4f$" % self.R2(), ha='center')
        return ax

    def filter(self, threshold):
        filtered = self.__class__(
            self.training_examples,
            self.testing_examples[self.score_array > threshold],
            self.model,
            self.scale,
            self.dilation,
            self.weighted)
        return filtered



from glypy.io import iupac, glycoct
from glypy import Glycan

from ms_deisotope.output import mgf, mzml, ProcessedMSFileLoader
from ms_deisotope.data_source import ScanBunch

from  sqlalchemy.orm import aliased
from glycresoft import serialize
from glycresoft.task import TaskBase


def build_partial_subgraph_n_glycan(composition):
    composition = composition.clone()
    if composition["HexNAc"] < 2:
        raise ValueError("Not enough HexNAc present to extract N-glycan core")
    root = iupac.loads("?-D-Glcp2NAc").clone(prop_id=False)
    composition['HexNAc'] -= 1
    b2 = iupac.loads("b-D-Glcp2NAc").clone(prop_id=False)
    root.add_monosaccharide(b2, position=4, child_position=1)
    composition['HexNAc'] -= 1
    if composition['Hex'] < 3:
        raise ValueError("Not enough Hex present to extract N-glycan core")
    composition['Hex'] -= 3
    b3 = iupac.loads("b-D-Manp").clone()
    b2.add_monosaccharide(b3, position=4, child_position=1)
    b4 = iupac.loads("a-D-Manp").clone()
    b3.add_monosaccharide(b4, position=3, child_position=1)
    b5 = iupac.loads("a-D-Manp").clone()
    b3.add_monosaccharide(b5, position=6, child_position=1)
    subgraph = Glycan(root, index_method=None)
    return composition, subgraph


def to_glycoct_partial_n_glycan(composition):
    a, b = build_partial_subgraph_n_glycan(composition)
    writer = glycoct.OrderRespectingGlycoCTWriter(b)
    writer.handle_glycan(writer.structure)
    writer.add_glycan_composition(a)
    return writer.buffer.getvalue()


class GlycoCTCompositionListExporter(TaskBase):
    def __init__(self, outstream, glycan_composition_iterable):
        self.outstream = outstream
        self.glycan_composition_iterable = glycan_composition_iterable

    def run(self):
        for glycan_composition in self.glycan_composition_iterable:
            text = to_glycoct_partial_n_glycan(glycan_composition)
            self.outstream.write(text)
            self.outstream.write("\n")


class AnnotatedMGFSerializer(mgf.MGFSerializer):

    def write_header(self, header_dict):
        super(AnnotatedMGFSerializer, self).write_header(header_dict)
        self.add_parameter("precursor_defaulted", header_dict['defaulted'])
        self.add_parameter("activation_method", header_dict['precursor_activation_method'])
        self.add_parameter("activation_energy", header_dict['precursor_activation_energy'])
        self.add_parameter("analyzers", header_dict.get('analyzers'))
        self.add_parameter("scan_id", header_dict['id'])
        self.add_parameter("precursor_scan_id", header_dict['precursor_scan_id'])
        for key, value in header_dict['annotations'].items():
            self.add_parameter(key, value)


class TrainingMGFExporterBase(TaskBase):
    def __init__(self, outstream, spectrum_match_iterable):
        self.outstream = outstream
        self.spectrum_match_iterable = spectrum_match_iterable

    def prepare_scan(self, spectrum_match):
        scan = spectrum_match.scan
        scan = scan.clone()
        scan.annotations['structure'] = spectrum_match.target
        # try:
        #     scan.annotations['protein_name'] = spectrum_match.target.protein.name
        # except AttributeError:
        #     pass
        try:
            scan.annotations['mass_shift'] = spectrum_match.mass_shift.name
        except AttributeError:
            pass
        scan.annotations['ms2_score'] = spectrum_match.score
        try:
            scan.annotations['q_value'] = spectrum_match.q_value
        except AttributeError:
            pass
        try:
            score_set = spectrum_match.score_set
            scan.annotations['ms2_peptide_score'] = score_set.peptide_score
            scan.annotations['ms2_glycan_score'] = score_set.glycan_score
            scan.annotations['ms2_glycan_coverage'] = score_set.glycan_coverage
            q_value_set = spectrum_match.q_value_set
            scan.annotations['q_value'] = q_value_set.total_q_value
            scan.annotations['glycan_q_value'] = q_value_set.glycan_q_value
            scan.annotations['peptide_q_value'] = q_value_set.peptide_q_value
            scan.annotations['glycopeptide_q_value'] = q_value_set.glycopeptide_q_value
        except AttributeError as err:
            print(err)
        return scan

    def run(self):
        writer = AnnotatedMGFSerializer(self.outstream)
        for spectrum_match in self.spectrum_match_iterable:
            scan = self.prepare_scan(spectrum_match)
            writer.save(ScanBunch(None, [scan]))
        writer.complete()


class TrainingMGFExporter(TrainingMGFExporterBase):
    def __init__(self, outstream, spectrum_match_iterable, mzml_path):
        super(TrainingMGFExporter, self).__init__(outstream, spectrum_match_iterable)
        self.scan_loader = ProcessedMSFileLoader(mzml_path)

    def prepare_scan(self, spectrum_match):
        scan = self.scan_loader.get_scan_by_id(spectrum_match.scan.scan_id)
        spectrum_match.scan = scan.clone()
        scan = super(TrainingMGFExporter, self).prepare_scan(spectrum_match)
        return scan

    @classmethod
    def _from_analysis_id_query(cls, database_connection, analysis_id, threshold=0.0):

        A = aliased(serialize.GlycopeptideSpectrumMatch)
        q = database_connection.query(A).filter(
            A.is_best_match,
            A.analysis_id == analysis_id)
        # B = aliased(serialize.GlycopeptideSpectrumMatch)
        # qinner = database_connection.query(
        #     B.id.label('inner_id'), serialize.func.max(B.score).label("inner_score"),
        #     B.scan_id.label("inner_scan_id")).group_by(B.scan_id).selectable

        # q = database_connection.query(A).join(
        #     qinner, qinner.c.inner_id == A.id and A.score == qinner.inner_score).filter(
        #     A.analysis_id == analysis_id)

        if threshold is not None:
            q = q.filter(A.score >= threshold)
        for gsm in q:
            yield gsm.convert()

    @classmethod
    def from_analysis(cls, database_connection, analysis_id, outstream, scan_loader_path=None, threshold=0.0):
        if scan_loader_path is None:
            analysis = database_connection.query(serialize.Analysis).get(analysis_id)
            scan_loader_path = analysis.parameters['sample_path']
        iterable = cls._from_analysis_id_query(database_connection, analysis_id, threshold=threshold)
        return cls(outstream, iterable, scan_loader_path)



from multiprocessing import Manager as IPCManager

from glycresoft.chromatogram_tree.chromatogram import GlycopeptideChromatogram
from glycresoft.chromatogram_tree import Unmodified
from glycresoft.task import TaskBase

from glycresoft.structure import (
    CachingGlycopeptideParser,
    SequenceReversingCachingGlycopeptideParser)


from ..target_decoy import GroupwiseTargetDecoyAnalyzer, TargetDecoySet
from .core_search import GlycanCombinationRecord, GlycanFilteringPeptideMassEstimator

from ..temp_store import TempFileManager, SpectrumMatchStore
from ..chromatogram_mapping import ChromatogramMSMSMapper
from ..workflow import (format_identification_batch, chunkiter, SearchEngineBase)

from ..oxonium_ions import OxoniumFilterReport, OxoniumFilterState

from .matcher import (
    TargetDecoyInterleavingGlycopeptideMatcher,
    ComparisonGlycopeptideMatcher)


class GlycopeptideResolver(object):
    def __init__(self, database, parser=None):
        if parser is None:
            parser = CachingGlycopeptideParser()
        self.database = database
        self.parser = parser
        self.cache = dict()

    def resolve(self, id):
        try:
            return self.cache[id]
        except KeyError:
            record = self.database.get_record(id)
            structure = self.parser(record)
            self.cache[id] = structure
            return structure

    def __call__(self, id):
        return self.resolve(id)


class GlycopeptideDatabaseSearchIdentifier(SearchEngineBase):
    def __init__(self, tandem_scans, scorer_type, structure_database, scan_id_to_rt=lambda x: x,
                 minimum_oxonium_ratio=0.05, scan_transformer=lambda x: x, mass_shifts=None,
                 n_processes=5, file_manager=None, use_peptide_mass_filter=True,
                 probing_range_for_missing_precursors=3, trust_precursor_fits=True,
                 permute_decoy_glycans=False):
        if file_manager is None:
            file_manager = TempFileManager()
        elif isinstance(file_manager, str):
            file_manager = TempFileManager(file_manager)
        if mass_shifts is None:
            mass_shifts = []
        if Unmodified not in mass_shifts:
            mass_shifts = [Unmodified] + mass_shifts
        self.tandem_scans = sorted(
            tandem_scans, key=lambda x: x.precursor_information.extracted_neutral_mass, reverse=True)
        self.scorer_type = scorer_type
        self.structure_database = structure_database
        self.scan_id_to_rt = scan_id_to_rt
        self.minimum_oxonium_ratio = minimum_oxonium_ratio
        self.mass_shifts = mass_shifts
        self.permute_decoy_glycans = permute_decoy_glycans

        self.probing_range_for_missing_precursors = probing_range_for_missing_precursors
        self.trust_precursor_fits = trust_precursor_fits

        self.use_peptide_mass_filter = use_peptide_mass_filter
        self._peptide_mass_filter = None

        self.scan_transformer = scan_transformer

        self.n_processes = n_processes
        self.ipc_manager = IPCManager()

        self.file_manager = file_manager
        self.spectrum_match_store = SpectrumMatchStore(self.file_manager)

    def _make_evaluator(self, bunch):
        evaluator = TargetDecoyInterleavingGlycopeptideMatcher(
            bunch, self.scorer_type, self.structure_database,
            minimum_oxonium_ratio=self.minimum_oxonium_ratio,
            n_processes=self.n_processes,
            ipc_manager=self.ipc_manager,
            mass_shifts=self.mass_shifts,
            peptide_mass_filter=self._peptide_mass_filter,
            probing_range_for_missing_precursors=self.probing_range_for_missing_precursors,
            trust_precursor_fits=self.trust_precursor_fits,
            permute_decoy_glycans=self.permute_decoy_glycans)
        return evaluator

    def prepare_scan_set(self, scan_set):
        if hasattr(scan_set[0], 'convert'):
            out = []
            # Account for cases where the scan may be mentioned in the index, but
            # not actually present in the MS data
            for o in scan_set:
                try:
                    scan = (self.scorer_type.load_peaks(o))
                    if len(scan.deconvoluted_peak_set) > 0:
                        out.append(scan)
                except KeyError:
                    self.log("Missing Scan: %s" % (o.id,))
            scan_set = out
        out = []
        unconfirmed_precursors = []
        for scan in scan_set:
            try:
                scan.deconvoluted_peak_set = self.scan_transformer(
                    scan.deconvoluted_peak_set)
                if len(scan.deconvoluted_peak_set) > 0:
                    if scan.precursor_information.defaulted:
                        unconfirmed_precursors.append(scan)
                    else:
                        out.append(scan)
            except AttributeError:
                self.log("Missing Scan: %s" % (scan.id,))
                continue
        return out, unconfirmed_precursors

    def _make_peptide_mass_filter(self, error_tolerance=1e-5):
        hypothesis_id = self.structure_database.hypothesis_id
        glycan_combination_list = GlycanCombinationRecord.from_hypothesis(
            self.structure_database.session, hypothesis_id)
        if len(glycan_combination_list) == 0:
            self.log("No glycan combinations were found")
            raise ValueError("No glycan combinations were found")
        peptide_filter = GlycanFilteringPeptideMassEstimator(
            glycan_combination_list, product_error_tolerance=error_tolerance)
        return peptide_filter

    def format_work_batch(self, bunch, count, total):
        ratio = "%d/%d (%0.3f%%)" % (count, total, (count * 100. / total))
        info = bunch[0].precursor_information
        try:
            try:
                precursor = info.precursor
                if hasattr(precursor, "scan_id"):
                    name = precursor.scan_id
                else:
                    name = precursor.id
            except (KeyError, AttributeError):
                if hasattr(bunch[0], "scan_id"):
                    name = bunch[0].scan_id
                else:
                    name = bunch[0].id
        except Exception:
            name = ""

        if isinstance(info.charge, (int, float)):
            batch_header = "%s: %f (%s%r)" % (
                name, info.neutral_mass, "+" if info.charge > 0 else "-", abs(
                    info.charge))
        else:
            batch_header = "%s: %f (%s)" % (
                name, info.neutral_mass, "?")
        return "Begin Batch", batch_header, ratio

    def search(self, precursor_error_tolerance=1e-5, simplify=True, batch_size=500, *args, **kwargs):
        target_hits = self.spectrum_match_store.writer("targets")
        decoy_hits = self.spectrum_match_store.writer("decoys")

        total = len(self.tandem_scans)
        count = 0

        if self.use_peptide_mass_filter:
            self._peptide_mass_filter = self._make_peptide_mass_filter(
                kwargs.get("error_tolerance", 1e-5))
        oxonium_report = OxoniumFilterReport()
        self.log("Writing Matches To %r" % (self.file_manager,))
        for scan_collection in chunkiter(self.tandem_scans, batch_size):
            count += len(scan_collection)
            for item in self.format_work_batch(scan_collection, count, total):
                self.log("... %s" % item)
            scan_collection, unconfirmed_precursors = self.prepare_scan_set(scan_collection)
            self.log("... %d Unconfirmed Precursor Spectra" % (len(unconfirmed_precursors,)))
            self.log("... Spectra Extracted")
            # TODO: handle unconfirmed_precursors differently here?
            evaluator = self._make_evaluator(scan_collection + unconfirmed_precursors)
            t, d = evaluator.score_all(
                precursor_error_tolerance=precursor_error_tolerance,
                simplify=simplify, *args, **kwargs)
            self.log("... Spectra Searched")
            target_hits.extend(o for o in t if o.score > 0.5)
            decoy_hits.extend(o for o in d if o.score > 0.5)
            t = sorted(t, key=lambda x: x.score, reverse=True)
            self.log("...... Total Matches So Far: %d Targets, %d Decoys\n%s" % (
                len(target_hits), len(decoy_hits), format_identification_batch(t, 10)))

            # clear these lists as they may be quite large and we don't need them around for the
            # next iteration
            t = []
            d = []
            oxonium_report.extend(evaluator.oxonium_ion_report)

        self.log('Search Done')
        target_hits.close()
        decoy_hits.close()
        self._clear_database_cache()

        self.log("Reloading Spectrum Matches")
        target_hits, decoy_hits = self._load_stored_matches(len(target_hits), len(decoy_hits))
        return TargetDecoySet(target_hits, decoy_hits)

    def _clear_database_cache(self):
        self.structure_database.clear_cache()

    def _load_stored_matches(self, target_count, decoy_count):
        target_resolver = GlycopeptideResolver(self.structure_database, CachingGlycopeptideParser(int(1e6)))
        decoy_resolver = GlycopeptideResolver(self.structure_database, SequenceReversingCachingGlycopeptideParser(int(1e6)))

        loaded_target_hits = []
        for i, solset in enumerate(self.spectrum_match_store.reader("targets", target_resolver)):
            if i % 5000 == 0:
                self.log("Loaded %d/%d Targets (%0.3g%%)" % (i, target_count, (100. * i / target_count)))
            loaded_target_hits.append(solset)
        loaded_decoy_hits = []
        for i, solset in enumerate(self.spectrum_match_store.reader("decoys", decoy_resolver)):
            if i % 5000 == 0:
                self.log("Loaded %d/%d Decoys (%0.3g%%)" % (i, decoy_count, (100. * i / decoy_count)))
            loaded_decoy_hits.append(solset)
        return TargetDecoySet(loaded_target_hits, loaded_decoy_hits)

    def estimate_fdr(self, target_hits, decoy_hits, with_pit=False, *args, **kwargs):
        self.log("Running Target Decoy Analysis with %d targets and %d decoys" % (
            len(target_hits), len(decoy_hits)))

        def over_10_aa(x):
            return len(x.target) >= 10

        def under_10_aa(x):
            return len(x.target) < 10

        grouping_fns = [over_10_aa, under_10_aa]

        tda = GroupwiseTargetDecoyAnalyzer(
            [x.best_solution() for x in target_hits],
            [x.best_solution() for x in decoy_hits], *args, with_pit=with_pit,
            grouping_functions=grouping_fns,
            grouping_labels=["Long Peptide", "Short Peptide"],
            **kwargs)
        tda.q_values()
        for sol in target_hits:
            for hit in sol:
                tda.score(hit)
            sol.q_value = sol.best_solution().q_value
        for sol in decoy_hits:
            for hit in sol:
                tda.score(hit)
            sol.q_value = sol.best_solution().q_value
        return tda

    def map_to_chromatograms(self, chromatograms, tandem_identifications,
                             precursor_error_tolerance=1e-5, threshold_fn=lambda x: x.q_value < 0.05,
                             entity_chromatogram_type=GlycopeptideChromatogram):
        self.log("Mapping MS/MS Identifications onto Chromatograms")
        self.log("%d Chromatograms" % len(chromatograms))
        # if len(chromatograms) == 0:
        #     self.log("No Chromatograms Extracted!")
        #     return chromatograms, tandem_identifications
        mapper = ChromatogramMSMSMapper(
            chromatograms, precursor_error_tolerance, self.scan_id_to_rt)
        self.log("Assigning Solutions")
        mapper.assign_solutions_to_chromatograms(tandem_identifications)
        self.log("Distributing Orphan Spectrum Matches")
        mapper.distribute_orphans(threshold_fn=threshold_fn)
        self.log("Selecting Most Representative Matches")
        mapper.assign_entities(threshold_fn, entity_chromatogram_type=entity_chromatogram_type)
        return mapper.chromatograms, mapper.orphans


class GlycopeptideDatabaseSearchComparer(GlycopeptideDatabaseSearchIdentifier):
    def __init__(self, tandem_scans, scorer_type, target_database, decoy_database, scan_id_to_rt=lambda x: x,
                 minimum_oxonium_ratio=0.05, scan_transformer=lambda x: x, mass_shifts=None,
                 n_processes=5, file_manager=None, use_peptide_mass_filter=True,
                 probing_range_for_missing_precursors=3, trust_precursor_fits=True,
                 permute_decoy_glycans=False):
        self.target_database = target_database
        self.decoy_database = decoy_database
        super(GlycopeptideDatabaseSearchComparer, self).__init__(
            tandem_scans, scorer_type, self.target_database, scan_id_to_rt,
            minimum_oxonium_ratio, scan_transformer, mass_shifts, n_processes,
            file_manager, use_peptide_mass_filter,
            probing_range_for_missing_precursors=probing_range_for_missing_precursors,
            trust_precursor_fits=trust_precursor_fits, permute_decoy_glycans=permute_decoy_glycans)

    def _clear_database_cache(self):
        self.target_database.clear_cache()
        self.decoy_database.clear_cache()

    def _make_evaluator(self, bunch):
        evaluator = ComparisonGlycopeptideMatcher(
            bunch, self.scorer_type,
            target_structure_database=self.target_database,
            decoy_structure_database=self.decoy_database,
            minimum_oxonium_ratio=self.minimum_oxonium_ratio,
            n_processes=self.n_processes,
            ipc_manager=self.ipc_manager,
            mass_shifts=self.mass_shifts,
            peptide_mass_filter=self._peptide_mass_filter,
            probing_range_for_missing_precursors=self.probing_range_for_missing_precursors,
            trust_precursor_fits=self.trust_precursor_fits,
            permute_decoy_glycans=self.permute_decoy_glycans)
        return evaluator

    def estimate_fdr(self, target_hits, decoy_hits, with_pit=False, *args, **kwargs):
        self.log("Running Target Decoy Analysis with %d targets and %d decoys" % (
            len(target_hits), len(decoy_hits)))

        database_ratio = float(len(self.target_database)) / len(self.decoy_database)

        def over_10_aa(x):
            return len(x.target) >= 10

        def under_10_aa(x):
            return len(x.target) < 10

        grouping_fns = [over_10_aa, under_10_aa]

        tda = GroupwiseTargetDecoyAnalyzer(
            [x.best_solution() for x in target_hits],
            [x.best_solution() for x in decoy_hits], *args, with_pit=with_pit,
            database_ratio=database_ratio, grouping_functions=grouping_fns,
            grouping_labels=["Long Peptide", "Short Peptide"]
            **kwargs)

        tda.q_values()
        for sol in target_hits:
            for hit in sol:
                tda.score(hit)
            sol.q_value = sol.best_solution().q_value
        for sol in decoy_hits:
            for hit in sol:
                tda.score(hit)
            sol.q_value = sol.best_solution().q_value
        return tda

    def _load_stored_matches(self, target_count, decoy_count):
        target_resolver = GlycopeptideResolver(self.target_database, CachingGlycopeptideParser(int(1e6)))
        decoy_resolver = GlycopeptideResolver(self.decoy_database, CachingGlycopeptideParser(int(1e6)))

        loaded_target_hits = []
        for i, solset in enumerate(self.spectrum_match_store.reader("targets", target_resolver)):
            if i % 5000 == 0:
                self.log("Loaded %d/%d Targets (%0.3g%%)" % (i, target_count, (100. * i / target_count)))
            loaded_target_hits.append(solset)
        loaded_decoy_hits = []
        for i, solset in enumerate(self.spectrum_match_store.reader("decoys", decoy_resolver)):
            if i % 5000 == 0:
                self.log("Loaded %d/%d Decoys (%0.3g%%)" % (i, decoy_count, (100. * i / decoy_count)))
            loaded_decoy_hits.append(solset)
        return TargetDecoySet(loaded_target_hits, loaded_decoy_hits)


class ExclusiveGlycopeptideDatabaseSearchComparer(GlycopeptideDatabaseSearchComparer):
    def estimate_fdr(self, target_hits, decoy_hits, with_pit=False, *args, **kwargs):
        accepted_targets, accepted_decoys = self._find_best_match_for_each_scan(target_hits, decoy_hits)
        tda = super(ExclusiveGlycopeptideDatabaseSearchComparer, self).estimate_fdr(
            accepted_targets, accepted_decoys, with_pit=with_pit, *args, **kwargs)
        for sol in target_hits:
            for hit in sol:
                tda.score(hit)
            sol.q_value = sol.best_solution().q_value
        for sol in decoy_hits:
            for hit in sol:
                tda.score(hit)
            sol.q_value = sol.best_solution().q_value
        return tda

    def _find_best_match_for_each_scan(self, target_hits, decoy_hits):
        winning_targets = []
        winning_decoys = []

        target_map = {t.scan.id: t for t in target_hits}
        decoy_map = {t.scan.id: t for t in decoy_hits}
        scan_ids = set(target_map) | set(decoy_map)
        for scan_id in scan_ids:
            target_sol = target_map.get(scan_id)
            decoy_sol = decoy_map.get(scan_id)
            if target_sol is None:
                winning_decoys.append(decoy_sol)
            elif decoy_sol is None:
                winning_targets.append(target_sol)
            else:
                if target_sol.score == decoy_sol.score:
                    winning_targets.append(target_sol)
                    winning_decoys.append(decoy_sol)
                elif target_sol.score > decoy_sol.score:
                    winning_targets.append(target_sol)
                else:
                    winning_decoys.append(decoy_sol)
        return winning_targets, winning_decoys



import glypy

from glycresoft.structure import FragmentMatchMap
from glycresoft.tandem.spectrum_match import SpectrumMatcherBase

water = glypy.Composition("H2O")
water_mass = water.mass


class GlycanSpectrumMatcherBase(SpectrumMatcherBase):

    def _match_fragments(self, series, error_tolerance=2e-5, max_cleavages=2, include_neutral_losses=None):
        for fragment in self.target.fragments(series, max_cleavages=max_cleavages):
            peaks = self.spectrum.all_peaks_for(fragment.mass, error_tolerance)
            for peak in peaks:
                self.solution_map.add(peak, fragment)
            if include_neutral_losses:
                for loss, label in include_neutral_losses:
                    peaks = self.spectrum.all_peaks_for(
                        fragment.mass - loss.mass, error_tolerance)
                    if peaks:
                        fragment_loss = fragment.copy()
                        fragment_loss.mass -= loss.mass
                        fragment_loss.composition -= loss
                        fragment_loss.name += label
                    for peak in peaks:
                        self.solution_map.add(peak, fragment_loss)

    def match(self, error_tolerance=2e-5, *args, **kwargs):
        self.solution_map = FragmentMatchMap()
        include_neutral_losses = kwargs.get("include_neutral_losses", False)
        max_cleavages = kwargs.get("max_cleavages", 2)
        is_hcd = self.is_hcd()
        is_exd = self.is_exd()
        if include_neutral_losses and isinstance(include_neutral_losses, (int, bool)):
            include_neutral_losses = [(glypy.Composition("H2O"), "-H2O")]

        if is_hcd:
            self._match_fragments(
                "BY", error_tolerance, max_cleavages=max_cleavages,
                include_neutral_losses=include_neutral_losses)
        else:
            self._match_fragments(
                "ABCXYZ", error_tolerance, max_cleavages=max_cleavages,
                include_neutral_losses=include_neutral_losses)

    def _theoretical_mass(self):
        return self.target.mass()



class ctxstream(object):
    def __init__(self, stream):
        self.stream = stream

    def write(self, string):
        self.stream.write(string)

    def flush(self):
        self.stream.flush()

    def __getattr__(self, attrib: str):
        attr = getattr(self.stream, attrib)
        return attr

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False



from .space import (n_glycan_distance, composition_distance, CompositionSpace)

from .rule import (
    CompositionRuleBase, CompositionExpressionRule, CompositionRangeRule,
    CompositionRatioRule, CompositionRuleClassifier)

from .neighborhood import (
    NeighborhoodCollection, NeighborhoodWalker, make_n_glycan_neighborhoods,
    make_adjacency_neighborhoods, make_mammalian_n_glycan_neighborhoods)

from .graph import (
    CompositionNormalizer, CompositionGraph, CompositionGraphNode,
    CompositionGraphEdge, EdgeSet, DijkstraPathFinder,
    GraphReader, GraphWriter, dump, load, normalize_composition)

__all__ = [
    "n_glycan_distance", "composition_distance", "CompositionSpace",

    "CompositionRuleBase", "CompositionExpressionRule", "CompositionRangeRule",
    "CompositionRatioRule", "CompositionRuleClassifier",

    "NeighborhoodCollection", "NeighborhoodWalker", "make_n_glycan_neighborhoods",
    "make_adjacency_neighborhoods", "make_mammalian_n_glycan_neighborhoods",

    "CompositionNormalizer", "CompositionGraph", "CompositionGraphNode",
    "CompositionGraphEdge", "EdgeSet", "DijkstraPathFinder",
    "GraphReader", "GraphWriter", "dump",
    "load", "normalize_composition",
]



from .base import ArtistBase

from .chromatogram_artist import (
    ChromatogramArtist, SmoothingChromatogramArtist,
    ChargeSeparatingChromatogramArtist, ChargeSeparatingSmoothingChromatogramArtist,
    NGlycanChromatogramColorizer, LabelProducer, NGlycanLabelProducer, n_glycan_labeler,
    AbundantLabeler, n_glycan_colorizer)

from .entity_bar_chart import (
    EntitySummaryBarChartArtist, AggregatedAbundanceArtist,
    BundledGlycanComposition, ScoreBarArtist)

from .colors import (
    ColorMapper)

from .glycan_visual_classification import (
    GlycanCompositionOrderer, GlycanCompositionClassifierColorizer,
    NGlycanCompositionColorizer, NGlycanCompositionOrderer,
    GlycanLabelTransformer)

from .plot_glycoforms import GlycoformLayout

from . import sequence_fragment_logo
from .sequence_fragment_logo import glycopeptide_match_logo

from .utils import figax, figure


__all__ = [
    "ChromatogramArtist", "SmoothingChromatogramArtist",
    "ChargeSeparatingChromatogramArtist", "ChargeSeparatingSmoothingChromatogramArtist",
    "NGlycanChromatogramColorizer", "LabelProducer", "NGlycanLabelProducer", "n_glycan_labeler",
    "AbundantLabeler", "ArtistBase",
    "n_glycan_colorizer", "ColorMapper",
    "EntitySummaryBarChartArtist", "AggregatedAbundanceArtist",
    "BundledGlycanComposition", "ScoreBarArtist",
    "GlycanCompositionOrderer", "GlycanCompositionClassifierColorizer",
    "NGlycanCompositionColorizer", "NGlycanCompositionOrderer",
    "GlycanLabelTransformer", "GlycoformLayout",
    "sequence_fragment_logo", "glycopeptide_match_logo",
    "figax", "figure",
]



import unittest
import tempfile


from glycresoft import serialize

from glycresoft.database.prebuilt import biosynthesis_human_n_linked


class GlycanCombinatoricsTests(unittest.TestCase):
    def setup_tempfile(self):
        file_name = tempfile.mktemp() + '.tmp'
        open(file_name, 'w').write('')
        return file_name

    def clear_file(self, path):
        open(path, 'wb')

    def setup_compositions(self, database_path, **kwargs):
        builder = biosynthesis_human_n_linked.BiosynthesisHumanNGlycansBuilder()
        result = builder.build(database_path, **kwargs)
        return result

    def test_prebuilt_native(self):
        path = self.setup_tempfile()
        hypothesis_builder = self.setup_compositions(path)
        glycan_count = hypothesis_builder.query(serialize.GlycanComposition).count()
        self.assertEqual(448, glycan_count)
        hypothesis_builder.close()
        self.clear_file(path)

    def test_prebuilt_permethylated(self):
        path = self.setup_tempfile()
        hypothesis_builder = self.setup_compositions(path, derivatization='methyl')
        glycan_count = hypothesis_builder.query(serialize.GlycanComposition).count()
        self.assertEqual(448, glycan_count)
        hypothesis_builder.close()
        self.clear_file(path)


if __name__ == '__main__':
    unittest.main()



# Excerpted from scipy.stats._kde and .stats._stats to allow releasing the GIL

from libc cimport math
cimport cython
cimport numpy as np

from numpy.math cimport PI
from numpy.math cimport INFINITY
from numpy.math cimport NAN
from numpy cimport ndarray, int64_t, float64_t, intp_t

import warnings

import numpy as np
import scipy.stats, scipy.special


ctypedef fused real:
    float
    double
    long double


@cython.wraparound(False)
@cython.initializedcheck(False)
@cython.cdivision(True)
@cython.boundscheck(False)
cdef inline int gaussian_kernel_estimate_inner(
    real[:, :] points_,
    real[:, :] values_,
    real[:, :] xi_,
    real[:, :] estimate,
    real[:, :] whitening,
    int n,
    int m,
    int d,
    int p,
) nogil:
    cdef:
        int i, j, k
        real residual, arg, norm

    # Evaluate the normalisation
    norm = math.pow((2 * PI), (- d / 2.))
    for i in range(d):
        norm *= whitening[i, i]

    for i in range(n):
        for j in range(m):
            arg = 0
            for k in range(d):
                residual = (points_[i, k] - xi_[j, k])
                arg += residual * residual

            arg = math.exp(-arg / 2.) * norm
            for k in range(p):
                estimate[j, k] += values_[i, k] * arg
    return 0


@cython.wraparound(False)
@cython.boundscheck(False)
def gaussian_kernel_estimate(points, values, xi, precision, dtype, real _=0):
    """
    def gaussian_kernel_estimate(points, real[:, :] values, xi, precision)
    Evaluate a multivariate Gaussian kernel estimate.
    Parameters
    ----------
    points : array_like with shape (n, d)
        Data points to estimate from in d dimensions.
    values : real[:, :] with shape (n, p)
        Multivariate values associated with the data points.
    xi : array_like with shape (m, d)
        Coordinates to evaluate the estimate at in d dimensions.
    precision : array_like with shape (d, d)
        Precision matrix for the Gaussian kernel.
    Returns
    -------
    estimate : double[:, :] with shape (m, p)
        Multivariate Gaussian kernel estimate evaluated at the input coordinates.
    """
    cdef:
        real[:, :] points_, xi_, values_, estimate, whitening
        int i, j, k
        int n, d, m, p
        real arg, residual, norm

    n = points.shape[0]
    d = points.shape[1]
    m = xi.shape[0]
    p = values.shape[1]

    if xi.shape[1] != d:
        raise ValueError("points and xi must have same trailing dim")
    if precision.shape[0] != d or precision.shape[1] != d:
        raise ValueError("precision matrix must match data dims")

    # Rescale the data
    whitening = np.linalg.cholesky(precision).astype(dtype, copy=False)
    points_ = np.dot(points, whitening).astype(dtype, copy=False)
    xi_ = np.dot(xi, whitening).astype(dtype, copy=False)
    values_ = values.astype(dtype, copy=False)

    # Create the result array and evaluate the weighted sum
    estimate = np.zeros((m, p), dtype)

    with nogil:
        gaussian_kernel_estimate_inner(
            points_,
            values_,
            xi_,
            estimate,
            whitening,
            n,
            m,
            d,
            p,
        )

    return np.asarray(estimate)


@cython.binding(True)
def evaluate_gaussian_kde(self, points):
    """Evaluate the estimated pdf on a set of points.
    Parameters
    ----------
    points : (# of dimensions, # of points)-array
        Alternatively, a (# of dimensions,) vector can be passed in and
        treated as a single point.
    Returns
    -------
    values : (# of points,)-array
        The values at each point.
    Raises
    ------
    ValueError : if the dimensionality of the input points is different than
                    the dimensionality of the KDE.
    """
    points = np.atleast_2d(np.asarray(points))

    d, m = points.shape
    if d != self.d:
        if d == 1 and m == self.d:
            # points was passed in as a row vector
            points = np.reshape(points, (self.d, 1))
            m = 1
        else:
            msg = "points have dimension %s, dataset has dimension %s" % (d,
                self.d)
            raise ValueError(msg)

    output_dtype = np.common_type(self.covariance, points)
    itemsize = np.dtype(output_dtype).itemsize
    if itemsize == 4:
        spec = 'float'
    elif itemsize == 8:
        spec = 'double'
    elif itemsize in (12, 16):
        spec = 'long double'
    else:
        raise TypeError('%s has unexpected item size %d' %
                        (output_dtype, itemsize))
    result = gaussian_kernel_estimate[spec](self.dataset.T, self.weights[:, None],
                                            points.T, self.inv_cov, output_dtype)
    return result[:, 0]